# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import shutil
import filecmp
from iva import kraken

modules_dir = os.path.dirname(os.path.abspath(kraken.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestKraken(unittest.TestCase):
    def setUp(self):
        self.db = kraken.Database(os.path.join(data_dir, 'kraken_test.db'))


    def test_count_cds_from_embl(self):
        '''test count_cds_from_embl'''
        for i in range(3):
            infile = os.path.join(data_dir, 'kraken_count_cds_from_embl.' + str(i))
            self.assertEqual(i, kraken.Database.count_cds_from_embl(infile))


    def test_get_parent_taxons(self):
        '''test _get_parent_taxons'''
        taxons = set(['1', '9', '13'])
        self.db._get_parent_taxons(taxons)
        expected = {
            '1': '1',
            '9': '32199',
            '13': '203488'
        }
        self.assertEqual(expected, self.db.taxon_to_parent)


    def test_load_extra_ref_info(self):
        '''test _load_extra_ref_info'''
        self.db.extra_refs_file = os.path.join(data_dir, 'kraken_test.extra_ids')
        self.db._load_extra_ref_info()
        expected = {
            2000000000: {
                'genbank_ids': ['10', '11', '12', '13'],
                'new_gis': [4000000000, 4000000001, 4000000002, 4000000003]
            },
            2000000001: {
                'genbank_ids': ['21'],
                'new_gis': [4000000004]
            },
            2000000002: {
                'genbank_ids': ['42', '43'],
                'new_gis': [4000000005, 4000000006]
            }
        }
        self.assertEqual(expected, self.db.extra_refs)


    def test_genbank_to_taxon_and_gi(self):
        '''test _genbank_to_taxon_and_gi'''
        expected = '43', '42'
        gb_file = os.path.join(data_dir, 'kraken_test.genbank_to_taxon_and_gi.gb')
        self.assertEqual(expected, self.db._genbank_to_taxon_and_gi(gb_file))


    def test_replace_fasta_header(self):
        '''test _replace_fasta_header'''
        tmp = 'tmp.test.fa'
        before = os.path.join(data_dir, 'kraken_test.replace_fasta_header.before.fa')
        after = os.path.join(data_dir, 'kraken_test.replace_fasta_header.after.fa')
        shutil.copyfile(before, tmp)
        self.db._replace_fasta_header(tmp, 'after')
        self.assertTrue(filecmp.cmp(tmp, after))
        os.unlink(tmp)


    def test_append_to_file(self):
        '''test _append_to_file'''
        tmp = 'tmp.test'
        before = os.path.join(data_dir, 'kraken_test.append_to_file.before')
        after = os.path.join(data_dir, 'kraken_test.append_to_file.after')
        shutil.copyfile(before, tmp)
        self.db._append_to_file(tmp, '42')
        self.assertTrue(filecmp.cmp(tmp, after))
        os.unlink(tmp)


    def test_species_to_dir(self):
        '''test species_to_dir'''
        self.assertEqual('a_b_c_d_e_f', self.db._species_to_embl_dir(r''' a\b/c(d)e.f-'''))
        self.assertEqual('added.12345', self.db._species_to_embl_dir('added.12345'))


    def test_get_most_common_species_dir(self):
        '''Test get_most_common_species'''
        report = os.path.join(data_dir, 'kraken_test.report')
        expected = os.path.join(self.db.embl_root, 'Human_immunodeficiency_virus_1')
        self.assertEqual(expected, self.db._get_most_common_species_dir(report))




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import os
import pysam
import tempfile
import shutil
from iva import contig, mapping, seed, mummer, graph, edge, common
import pyfastaq

class Assembly:
    def __init__(self, contigs_file=None, map_index_k=15, map_index_s=3, threads=1, kmc_threads=1, max_insert=800, map_minid=0.5, min_clip=3, ext_min_cov=5, ext_min_ratio=2, ext_bases=100, verbose=0, seed_min_cov=5, seed_min_ratio=10, seed_min_kmer_count=200, seed_max_kmer_count=1000000000, seed_start_length=None, seed_stop_length=500, seed_overlap_length=None, make_new_seeds=False, contig_iter_trim=10, seed_ext_max_bases=50, max_contigs=50, clean=True, strand_bias=0):
        self.contigs = {}
        self.contig_lengths = {}
        self.map_index_k = map_index_k
        self.map_index_s = map_index_s
        self.threads = threads
        self.kmc_threads = kmc_threads
        self.max_insert = max_insert
        self.map_minid = map_minid
        self.min_clip = min_clip
        self.ext_min_cov = ext_min_cov
        self.ext_min_ratio = ext_min_ratio
        self.ext_bases = ext_bases
        self.verbose = verbose
        self.clean = clean
        self.make_new_seeds = make_new_seeds
        self.seed_start_length = seed_start_length
        self.seed_stop_length = seed_stop_length
        self.seed_min_kmer_count = seed_min_kmer_count
        self.seed_max_kmer_count = seed_max_kmer_count
        self.seed_ext_max_bases = seed_ext_max_bases
        self.seed_overlap_length = seed_overlap_length
        self.seed_min_cov = seed_min_cov
        self.seed_min_ratio = seed_min_ratio
        self.contig_iter_trim = contig_iter_trim
        self.max_contigs = max_contigs
        self.strand_bias = strand_bias
        self.contigs_trimmed_for_strand_bias = set()
        self.used_seeds = set()

        if contigs_file is None:
            self.make_new_seeds = True
        else:
            contigs = {}
            pyfastaq.tasks.file_to_dict(contigs_file, contigs)
            for ctg in contigs:
                self._add_contig(contigs[ctg])


    def _add_contig(self, ctg, min_length=1):
        if len(ctg) < min_length:
            return
        assert ctg.id not in self.contigs
        assert len(ctg) > 0
        self.contigs[ctg.id] = contig.Contig(ctg, verbose=self.verbose)
        self.contig_lengths[ctg.id] = [[len(self.contigs[ctg.id]), 0, 0]]


    def write_contigs_to_file(self, filename, min_length=None, do_not_write=None, only_write=None, biggest_first=False, order_by_orfs=False, prefix=None):
        if do_not_write is None:
            do_not_write = set()
        if only_write is None:
            only_write = set()

        printed = 0
        if min_length is None:
            min_length = self.map_index_k + 1
        f = pyfastaq.utils.open_file_write(filename)
        if biggest_first:
            contig_names = self._contig_names_size_order(biggest_first=True)
        elif order_by_orfs:
            names = self._get_contig_order_by_orfs()
            contig_names = [x[0] for x in names]
            contig_revcomp = [x[1] for x in names]
        else:
            contig_names = sorted(list(self.contigs.keys()))

        for i in range(len(contig_names)):
            name = contig_names[i]

            if len(self.contigs[name]) >= min_length and name not in do_not_write and (name in only_write or len(only_write)==0):
                if order_by_orfs and contig_revcomp[i]:
                    self.contigs[name].fa.revcomp()

                if prefix is None:
                    print(self.contigs[name].fa, file=f)
                else:
                    printed += 1
                    self.contigs[name].fa.id = prefix + '.' + str(printed).zfill(5)
                    print(self.contigs[name].fa, file=f)
                    self.contigs[name].fa.id = name

                if order_by_orfs and contig_revcomp[i]:
                    self.contigs[name].fa.revcomp()

        pyfastaq.utils.close(f)


    def _get_contig_order_by_orfs(self, min_length=300):
        longest_orfs = []
        no_orfs = set()
        ordered_names = []
        for contig in self.contigs.values():
            orfs = contig.fa.all_orfs(min_length)
            reverse = False
            max_length = 0

            for coords, rev in orfs:
                if len(coords) > max_length:
                    max_length = len(coords)
                    reverse = rev

            if max_length > 0:
                longest_orfs.append((contig.fa.id, max_length, reverse))
            else:
                no_orfs.add((contig.fa.id, len(contig.fa)))

        all_in_size_order = self._contig_names_size_order(biggest_first=True)
        ordered_names = sorted(longest_orfs, key=lambda x:x[1], reverse=True)
        ordered_names = [(x[0], x[2]) for x in ordered_names]

        for t in sorted(no_orfs, key=lambda x:x[1], reverse=True):
            ordered_names.append((t[0], False))

        return ordered_names


    def _map_reads(self, fwd_reads, rev_reads, out_prefix, required_flag=None, exclude_flag=None, sort_reads=False, mate_ref=None, no_map_contigs=None):
        if no_map_contigs is None:
            no_map_contigs = set()
        if self.verbose:
            print('    map reads', fwd_reads, rev_reads, sep='\t')
        reference = out_prefix + '.ref.fa'
        self.write_contigs_to_file(reference, do_not_write=no_map_contigs)
        mapping.map_reads(fwd_reads, rev_reads, reference, out_prefix, index_k=self.map_index_k, index_s=self.map_index_s, threads=self.threads, max_insert=self.max_insert, minid=self.map_minid, verbose=self.verbose, required_flag=required_flag, sort=sort_reads, exclude_flag=exclude_flag)
        if self.clean:
            os.unlink(reference)
        os.unlink(reference + '.fai')


    def _extend_contigs_with_bam(self, bam_in, out_prefix=None, output_all_useful_reads=False):
        if out_prefix is not None:
            fa_out1 = pyfastaq.utils.open_file_write(out_prefix + '_1.fa')
            fa_out2 = pyfastaq.utils.open_file_write(out_prefix + '_2.fa')
        keep_read_types = set([mapping.CAN_EXTEND_LEFT, mapping.CAN_EXTEND_RIGHT, mapping.KEEP])
        if output_all_useful_reads:
            keep_read_types.add(mapping.BOTH_UNMAPPED)
        previous_sam = None
        left_seqs = []
        right_seqs = []
        sam_reader = pysam.Samfile(bam_in, "rb")

        for current_sam in sam_reader.fetch(until_eof=True):
            if previous_sam is None:
                previous_sam = current_sam
                continue

            previous_type, current_type = mapping.get_pair_type(previous_sam, current_sam, self._get_ref_length_sam_pair(sam_reader, previous_sam, current_sam), self.max_insert, min_clip=self.min_clip)

            for sam, sam_type in [(previous_sam, previous_type), (current_sam, current_type)]:
                if sam_type == mapping.CAN_EXTEND_LEFT:
                    name = mapping.get_ref_name(sam, sam_reader)
                    clipped = mapping.soft_clipped(sam)[0]
                    self.contigs[name].add_left_kmer(common.decode(sam.seq[:clipped]))
                elif sam_type == mapping.CAN_EXTEND_RIGHT:
                    name = mapping.get_ref_name(sam, sam_reader)
                    self.contigs[name].add_right_kmer(common.decode(sam.seq[sam.qend:]))

                if out_prefix is not None and sam_type in keep_read_types:
                    if sam.is_read1:
                        print(mapping.sam_to_fasta(sam), file=fa_out1)
                    else:
                        print(mapping.sam_to_fasta(sam), file=fa_out2)

            previous_sam = None

        if out_prefix is not None:
            pyfastaq.utils.close(fa_out1)
            pyfastaq.utils.close(fa_out2)
        total_bases_added = 0

        for ctg in self.contigs:
            left_length, right_length = self.contigs[ctg].extend(self.ext_min_cov, self.ext_min_ratio, self.ext_bases)
            if self.verbose:
                print('    extend contig ' +  ctg, 'new_length:' + str(len(self.contigs[ctg])), 'added_left:' + str(left_length), 'added_right:' + str(right_length), sep='\t')
            self.contig_lengths[ctg].append([len(self.contigs[ctg]), left_length, right_length])
            total_bases_added += left_length + right_length

        return total_bases_added


    def _trim_contig_for_strand_bias(self, bam, ctg_name):
        assert os.path.exists(bam)
        if ctg_name in self.contigs_trimmed_for_strand_bias:
            return
        ctg_length = len(self.contigs[ctg_name])
        fwd_cov = mapping.get_bam_region_coverage(bam, ctg_name, ctg_length)
        rev_cov = mapping.get_bam_region_coverage(bam, ctg_name, ctg_length, rev=True)
        first_good_base = 0
        while first_good_base < ctg_length:
            total_cov = fwd_cov[first_good_base] + rev_cov[first_good_base]
            if total_cov >= self.ext_min_cov and min(fwd_cov[first_good_base], rev_cov[first_good_base]) / total_cov >= self.strand_bias:
                break
            first_good_base += 1

        last_good_base = ctg_length - 1
        while last_good_base > first_good_base:
            total_cov = fwd_cov[last_good_base] + rev_cov[last_good_base]
            if total_cov >= self.ext_min_cov and min(fwd_cov[last_good_base], rev_cov[last_good_base]) / total_cov >= self.strand_bias:
                break
            last_good_base -= 1

        if self.verbose >= 2:
            print('Trimming strand biased ends of contig', ctg_name, '- good base range is', first_good_base + 1, 'to', last_good_base + 1, 'from', ctg_length, 'bases')
        self.contigs[ctg_name].fa.seq = self.contigs[ctg_name].fa.seq[first_good_base:last_good_base+1]


    def _good_intervals_from_strand_coverage(self, fwd_cov, rev_cov):
        assert len(fwd_cov) == len(rev_cov)
        good_intervals = []
        start = None
        cov_ok = False
        for i in range(len(fwd_cov)):
            total_cov = fwd_cov[i] + rev_cov[i]
            cov_ok = total_cov >= self.ext_min_cov and min(fwd_cov[i], rev_cov[i]) / total_cov >= self.strand_bias
            if cov_ok:
                if start is None:
                    start = i
            else:
                if start is not None:
                    good_intervals.append((start, i-1))
                start = None

        if cov_ok and start is not None:
            good_intervals.append((start, i-1))
        return good_intervals


    def _subcontigs_from_strand_bias(self, bam, ctg_name):
        ctg_length = len(self.contigs[ctg_name])
        fwd_cov = mapping.get_bam_region_coverage(bam, ctg_name, ctg_length)
        rev_cov = mapping.get_bam_region_coverage(bam, ctg_name, ctg_length, rev=True)
        good_intervals = self._good_intervals_from_strand_coverage(fwd_cov, rev_cov)
        new_contigs = []

        if len(good_intervals) == 1:
            self.contigs[ctg_name].fa.seq = self.contigs[ctg_name].fa.seq[good_intervals[0][0]:good_intervals[0][1]+1]
        elif len(good_intervals) > 1:
            for i in range(len(good_intervals)):
                start = good_intervals[i][0]
                end = good_intervals[i][1]
                if end - start + 1 >= 100:
                    new_contigs.append(pyfastaq.sequences.Fasta(ctg_name + '.' + str(i+1), self.contigs[ctg_name].fa[start:end+1]))

        return new_contigs


    def _trim_strand_biased_ends(self, reads_prefix, out_prefix=None, tag_as_trimmed=False, break_contigs=False):
        tmpdir = tempfile.mkdtemp(prefix='tmp.trim_strand_biased_ends.', dir=os.getcwd())
        tmp_prefix = os.path.join(tmpdir, 'out')
        sorted_bam = tmp_prefix + '.bam'
        unsorted_bam = tmp_prefix + '.unsorted.bam'
        original_map_minid = self.map_minid
        self.map_minid = 0.9
        self._map_reads(reads_prefix + '_1.fa', reads_prefix + '_2.fa', tmp_prefix, sort_reads=True)
        assert os.path.exists(sorted_bam)
        self.map_minid = original_map_minid
        new_contigs = []
        contigs_to_remove = set()
        for ctg in self.contigs:
            if break_contigs:
                subcontigs = self._subcontigs_from_strand_bias(sorted_bam, ctg)
                if len(subcontigs):
                    new_contigs.extend(subcontigs)
                    contigs_to_remove.add(ctg)
            elif ctg not in self.contigs_trimmed_for_strand_bias:
                self._trim_contig_for_strand_bias(sorted_bam, ctg)
                # contig could get completely trimmed so nothing left, in which
                # case, we need to remove it
                if len(self.contigs[ctg]) == 0:
                    contigs_to_remove.add(ctg)
                elif tag_as_trimmed:
                    self.contigs_trimmed_for_strand_bias.add(ctg)

        for ctg in contigs_to_remove:
            self._remove_contig(ctg)

        for ctg in new_contigs:
            self._add_contig(ctg, min_length=0.75 * self.self.seed_stop_length)


        if out_prefix is not None:
            mapping.bam_file_to_fasta_pair_files(unsorted_bam, out_prefix + '_1.fa', out_prefix + '_2.fa', remove_proper_pairs=True)
        shutil.rmtree(tmpdir)


    def trim_contigs(self, trim):
        for ctg in self.contigs:
            if self._contig_worth_extending(ctg):
                self.contigs[ctg].fa.trim(trim, trim)
                self.contig_lengths[ctg][-1][0] -= 2 * trim


    def _contig_worth_extending(self, name):
        if name in self.contigs_trimmed_for_strand_bias:
            return False
        return len(self.contig_lengths[name]) < 3 \
                  or self.contig_lengths[name][-1][0] > max([self.contig_lengths[name][x][0] for x in range(len(self.contig_lengths[name])-2)])


    def _worth_extending(self):
        for ctg in self.contigs:
            if self._contig_worth_extending(ctg):
                return True
        return False


    def _extend_with_reads(self, reads_prefix, out_prefix, no_map_contigs):
        tmpdir = tempfile.mkdtemp(prefix='tmp.extend.', dir=os.getcwd())
        tmp_prefix = os.path.join(tmpdir, 'reads')
        total_bases_added = 0

        for i in range(5):
            bam_prefix = out_prefix + '.' + str(i+1) + '.map'
            bam = bam_prefix +  '.bam'
            self._map_reads(reads_prefix + '_1.fa', reads_prefix + '_2.fa', bam_prefix, no_map_contigs=no_map_contigs)
            reads_prefix = tmp_prefix + '.' + str(i)
            bases_added = self._extend_contigs_with_bam(bam, out_prefix=reads_prefix)
            total_bases_added += bases_added
            if self.clean:
                os.unlink(bam)
            if bases_added < 0.2 * self.ext_bases:
                break

        shutil.rmtree(tmpdir)
        return total_bases_added


    def _read_pair_extension_iterations(self, reads_prefix, out_prefix, no_map_contigs=None):
        if no_map_contigs is None:
            no_map_contigs = set()
        assert(len(self.contigs) > len(no_map_contigs))
        if self.verbose:
            print('{:-^79}'.format(' ' + out_prefix + ' start extension subiteration 0001 '), flush=True)

        bases_added = self._extend_with_reads(reads_prefix, out_prefix + '.1', no_map_contigs)
        current_reads_prefix = reads_prefix

        if bases_added == 0:
            return True
        try_contig_trim = False
        i = 1

        while self._worth_extending() or try_contig_trim:
            i += 1
            if self.verbose:
                print('{:-^79}'.format(' ' + out_prefix + ' start extension subiteration ' + str(i).zfill(4) + ' '), flush=True)

            if i % 5 == 0:
                tmpdir = tempfile.mkdtemp(prefix='tmp.filter_reads.', dir=os.getcwd())
                tmp_prefix = os.path.join(tmpdir, 'out')
                bam = tmp_prefix + '.bam'
                original_map_minid = self.map_minid
                self.map_minid = 0.9
                self._map_reads(current_reads_prefix + '_1.fa', current_reads_prefix + '_2.fa', tmp_prefix)
                self.map_minid = original_map_minid
                filter_prefix = reads_prefix + '.subiter.' + str(i) + '.reads'
                mapping.bam_file_to_fasta_pair_files(bam, filter_prefix + '_1.fa', filter_prefix + '_2.fa', remove_proper_pairs=True)
                if current_reads_prefix != reads_prefix:
                    os.unlink(current_reads_prefix + '_1.fa')
                    os.unlink(current_reads_prefix + '_2.fa')
                current_reads_prefix = filter_prefix
                shutil.rmtree(tmpdir)

            iter_prefix = out_prefix + '.' + str(i)
            bases_added = self._extend_with_reads(current_reads_prefix, iter_prefix, no_map_contigs)

            if bases_added == 0:
                if not try_contig_trim:
                    if self.verbose:
                        print('    No bases added. Try trimming contigs')
                    self._trim_strand_biased_ends(reads_prefix, tag_as_trimmed=False)
                    if len(self.contigs) <= len(no_map_contigs):
                        if self.verbose:
                            print('       lost contigs during trimming. No more iterations')
                        return False
                    self.trim_contigs(self.contig_iter_trim)
                    try_contig_trim = True
                else:
                    if self.verbose:
                        print('    No bases added after trimming. No more iterations')
                    break
            else:
                try_contig_trim = False

        if current_reads_prefix != reads_prefix:
            os.unlink(current_reads_prefix + '_1.fa')
            os.unlink(current_reads_prefix + '_2.fa')
        return True


    def read_pair_extend(self, reads_prefix, out_prefix):
        assert(len(self.contigs))
        current_reads_prefix = reads_prefix
        i = 1
        new_seed_name = 'dummy'

        while 1:
            if self.verbose:
                print('{:_^79}'.format(' START ITERATION ' + str(i) + ' '), flush=True)
            self._read_pair_extension_iterations(current_reads_prefix, out_prefix + '.' + str(i))
            filtered_reads_prefix = out_prefix + '.' + str(i) + '.filtered'
            self._trim_strand_biased_ends(reads_prefix, tag_as_trimmed=True, out_prefix=filtered_reads_prefix)
            self._remove_contained_contigs(list(self.contigs.keys()))
            self._merge_overlapping_contigs(list(self.contigs.keys()))
            if reads_prefix != current_reads_prefix:
                os.unlink(current_reads_prefix + '_1.fa')
                os.unlink(current_reads_prefix + '_2.fa')
            current_reads_prefix = filtered_reads_prefix
            i += 1
            reads_left = os.path.getsize(current_reads_prefix + '_1.fa') > 0 and os.path.getsize(current_reads_prefix + '_2.fa') > 0

            if not self.make_new_seeds or new_seed_name is None or not self.make_new_seeds or len(self.contigs) >= self.max_contigs or not reads_left:
                if reads_prefix != current_reads_prefix:
                    os.unlink(current_reads_prefix + '_1.fa')
                    os.unlink(current_reads_prefix + '_2.fa')
                break

            if self.verbose:
                print('{:_^79}'.format(' Try making new seed '), flush=True)
            new_seed_name = self.add_new_seed_contig(current_reads_prefix + '_1.fa', current_reads_prefix + '_2.fa')

            if new_seed_name is None:
                if self.verbose:
                    print('Couldn\'t make new seed and extend it. Stopping assembly.')
                if len(self.contigs) == 0:
                    print('No contigs made.')
                    print('Read coverage may be too low, in which case try reducing --seed_min_kmer_cov, --ext_min_cov and --seed_ext_min_cov.')
                    print('Alternatively --max_insert could be incorrect, which is currently set to:', self.max_insert)
                if reads_prefix != current_reads_prefix:
                    os.unlink(current_reads_prefix + '_1.fa')
                    os.unlink(current_reads_prefix + '_2.fa')
                break



    def _run_nucmer(self, contigs_to_use=None):
        if contigs_to_use is None:
            contigs_to_use = set()
        if len(contigs_to_use) + len(self.contigs) <= 1:
            return []
        tmpdir = tempfile.mkdtemp(prefix='tmp.remove_self_contigs.', dir=os.getcwd())
        nucmer_out = os.path.join(tmpdir, 'nucmer.out')
        contigs_fasta = os.path.join(tmpdir, 'contigs.fa')
        self.write_contigs_to_file(contigs_fasta, only_write=contigs_to_use)
        mummer.run_nucmer(contigs_fasta, contigs_fasta, nucmer_out)
        hits = [hit for hit in mummer.file_reader(nucmer_out) if not hit.is_self_hit()]
        for hit in hits:
            hit.sort()
        hits = list(set(hits))
        shutil.rmtree(tmpdir)
        return hits


    def _remove_contained_contigs(self, contigs):
        if len(contigs) <= 1:
            return
        hits = self._run_nucmer(contigs_to_use=contigs)
        for contig in self._contig_names_size_order()[:-1]:
            if self._contig_contained_in_nucmer_hits(hits, contig, 95):
                hits = self._remove_contig_from_nucmer_hits(hits, contig)
                self._remove_contig(contig)
                contigs.remove(contig)


    def _coords_to_new_contig(self, coords_list):
        new_contig = pyfastaq.sequences.Fasta(coords_list[0][0], '')
        for name, coords, reverse in coords_list:
            assert name in self.contigs
            if reverse:
                seq = pyfastaq.sequences.Fasta('ni', self.contigs[name].fa.seq[coords.start:coords.end+1])
                seq.revcomp()
                new_contig.seq += seq.seq
            else:
                new_contig.seq += self.contigs[name].fa.seq[coords.start:coords.end+1]

        return new_contig


    def _merge_overlapping_contigs(self, contigs):
        if len(contigs) <= 1:
            return
        hits = self._run_nucmer(contigs)
        assembly_graph = graph.Graph(self, contigs=contigs)
        for hit in hits:
            e = hit.to_graph_edge()
            if e is not None:
                assembly_graph.add_edge(e)

        for connected_component in assembly_graph.connected_components():
            if len(connected_component) < 2:
                continue
            simple_path = assembly_graph.find_simple_path(connected_component)
            assert assembly_graph.simple_path_is_consistent(simple_path)
            if len(simple_path) > 1:
                simple_path = assembly_graph.remove_redundant_nodes_from_simple_path(simple_path)
                coords = assembly_graph.merged_coords_from_simple_nonredundant_path(simple_path)
                new_contig = self._coords_to_new_contig(coords)
                for name, x, y in coords:
                    self._remove_contig(name)
                self._add_contig(new_contig)


    def _contig_names_size_order(self, biggest_first=False):
        return sorted(self.contigs, key=lambda x:len(self.contigs[x]), reverse=biggest_first)


    def _contig_contained_in_nucmer_hits(self, hits, contig, min_percent):
        assert contig in self.contigs
        contig_length = len(self.contigs[contig])
        coords = []
        for hit in [hit for hit in hits if contig in [hit.qry_name, hit.ref_name] and hit.qry_name != hit.ref_name]:
            start = min(hit.qry_start, hit.qry_end)
            end = max(hit.qry_start, hit.qry_end)
            coords.append(pyfastaq.intervals.Interval(start, end))

        if len(coords) == 0:
            return False

        pyfastaq.intervals.merge_overlapping_in_list(coords)
        total_bases_matched = pyfastaq.intervals.length_sum_from_list(coords)
        return min_percent <= 100.0 * total_bases_matched / len(self.contigs[contig])


    def _remove_contig_from_nucmer_hits(self, hits, contig):
        return [x for x in hits if contig not in [x.ref_name, x.qry_name]]


    def _remove_contig(self, contig):
        if contig in self.contigs:
            del self.contigs[contig]
        if contig in self.contig_lengths:
            del self.contig_lengths[contig]
        if contig in self.contigs_trimmed_for_strand_bias:
            self.contigs_trimmed_for_strand_bias.remove(contig)


    def _get_ref_length(self, samfile, sam):
        if sam.is_unmapped:
            return None
        else:
            return len(self.contigs[mapping.get_ref_name(sam, samfile)])


    def _get_ref_length_sam_pair(self, samfile, sam1, sam2):
        len1 = self._get_ref_length(samfile, sam1)
        len2 = self._get_ref_length(samfile, sam2)
        if len1 == len2:
            return len1
        else:
            return None


    def _get_unmapped_pairs(self, reads1, reads2, out_prefix):
        self._map_reads(reads1, reads2, out_prefix, required_flag=12)
        mapping.bam_file_to_fasta_pair_files(out_prefix + '.bam', out_prefix + '_1.fa', out_prefix + '_2.fa')
        os.unlink(out_prefix + '.bam')


    def add_new_seed_contig(self, reads1, reads2, contig_name=None, max_attempts=10):
        if len(self.contigs):
            tmpdir = tempfile.mkdtemp(prefix='tmp.make_seed.', dir=os.getcwd())
            tmp_prefix = os.path.join(tmpdir, 'out')
            seed_reads1 = tmp_prefix + '_1.fa'
            seed_reads2 = tmp_prefix + '_2.fa'
            if contig_name is not None:
                self._map_reads(reads1, reads2, tmp_prefix, required_flag=5, exclude_flag=8, mate_ref=contig_name)
                mapping.bam_to_fasta(tmp_prefix + '.bam', seed_reads1)
                seed_reads2 = None
            else:
                self._get_unmapped_pairs(reads1, reads2, tmp_prefix)
        else:
            seed_reads1 = reads1
            seed_reads2 = reads2

        made_seed = False

        for i in range(max_attempts):
            s = seed.Seed(reads1=seed_reads1, reads2=seed_reads2, extend_length=self.seed_ext_max_bases, seed_length=self.seed_start_length, seed_min_count=self.seed_min_kmer_count, seed_max_count=self.seed_max_kmer_count, ext_min_cov=self.seed_min_cov, ext_min_ratio=self.seed_min_ratio, verbose=self.verbose, kmc_threads=self.kmc_threads, map_threads=self.threads, sequences_to_ignore=self.used_seeds, contigs_to_check=self.contigs)

            if s.seq is None or len(s.seq) == 0:
                break

            if self.seed_overlap_length is None:
                s.overlap_length = len(s.seq)
            else:
                s.overlap_length = self.seed_overlap_length
            s.extend(reads1, reads2, self.seed_stop_length)
            self.used_seeds.add(s.seq)

            if len(s.seq) >= 0.75 * self.seed_stop_length:
                made_seed = True
                break
            elif self.verbose:
                print("    Couldn't extend seed enough. That was attempt", i+1, 'of', max_attempts, flush=True)

        if len(self.contigs):
            shutil.rmtree(tmpdir)

        if not made_seed or len(s.seq) == 0:
            return None

        if self.verbose:
            print("    Extended seed OK.", flush=True)
        new_name = 'seeded.' + '1'.zfill(5)
        i = 1
        while new_name in self.contigs:
            i += 1
            new_name = 'seeded.' + str(i).zfill(5)

        self._add_contig(pyfastaq.sequences.Fasta(new_name, s.seq))
        return new_name




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import filecmp
import pysam
from iva import graph, assembly, edge
from pyfastaq import intervals

modules_dir = os.path.dirname(os.path.abspath(graph.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.asm = assembly.Assembly(contigs_file=os.path.join(data_dir, 'graph_test.contigs.fa'))
        self.g = graph.Graph(self.asm)


    def test_get_nodes(self):
        '''test get_nodes'''
        self.assertListEqual(['c1', 'c2', 'c3', 'c4', 'c5', 'c6'], self.g.get_nodes())


    def test_remove_node(self):
        '''test remove_node'''
        # TODO
        pass


    def test_add_and_get_edges(self):
        '''test adding and getting edges'''
        self.assertListEqual([], self.g._get_edges('not_in_graph', 'c1'))
        self.assertListEqual([], self.g._get_edges('not_in_graph', 'c2'))
        self.assertListEqual([], self.g._get_edges('c1', 'c2'))
        self.assertListEqual([], self.g._get_edges('c2', 'c1'))

        e1 = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e2 = edge.Edge('c1', 4, 43, 'c2', 60, 10)
        self.g.add_edge(e1)
        self.assertListEqual([e1], self.g._get_edges('c1', 'c2'))
        self.g.add_edge(e2)
        self.assertListEqual([e1, e2], self.g._get_edges('c1', 'c2'))

        with self.assertRaises(graph.Error):
            self.g.add_edge(edge.Edge('c7', 1, 10, 'c8', 42, 84))


    def test_degree(self):
        '''test degree'''
        self.assertEqual(0, self.g._degree('c1'))
        self.assertEqual(0, self.g._degree('c2'))
        e1 = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e2 = edge.Edge('c1', 4, 43, 'c2', 60, 10)
        self.g.add_edge(e1)
        self.assertEqual(1, self.g._degree('c1'))
        self.assertEqual(1, self.g._degree('c2'))
        self.g.add_edge(e2)
        self.assertEqual(2, self.g._degree('c1'))
        self.assertEqual(2, self.g._degree('c2'))


    def test_connected_components(self):
        '''test connected_components'''
        self.assertListEqual([['c1'], ['c2'], ['c3'], ['c4'], ['c5'], ['c6']], self.g.connected_components())
        self.g.add_edge(edge.Edge('c1', 42, 1, 'c2', 1, 10))
        self.assertListEqual([['c1', 'c2'], ['c3'], ['c4'], ['c5'], ['c6']], self.g.connected_components())
        self.g.add_edge(edge.Edge('c3', 42, 1, 'c2', 1, 10))
        self.assertListEqual([['c1', 'c2', 'c3'], ['c4'], ['c5'], ['c6']], self.g.connected_components())


    def test_find_simple_path(self):
        '''test find_simple_path'''
        self.assertListEqual([], self.g.find_simple_path(['c1']))
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.assertListEqual(['c1', 'c2'], self.g.find_simple_path(['c1', 'c2']))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertListEqual(['c1', 'c2', 'c3'], self.g.find_simple_path(['c1', 'c2', 'c3']))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertListEqual([], self.g.find_simple_path(['c1', 'c2', 'c3']))
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.assertListEqual([], self.g.find_simple_path(['c1', 'c2']))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.g.add_edge(edge.Edge('c3', 10, 100, 'c4', 1, 1000))
        self.assertListEqual(['c1', 'c2', 'c3', 'c4'], self.g.find_simple_path(['c1', 'c2', 'c3', 'c4']))
        self.g.add_edge(edge.Edge('c3', 10, 100, 'c5', 1, 1000))
        self.assertListEqual([], self.g.find_simple_path(['c1', 'c2', 'c3']))


    def test_remove_middle_node(self):
        '''test remove_middle_node'''
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 100, 142, 'c3', 1, 42))
        self.assertFalse(self.g._remove_middle_node('c1', 'c2', 'c3'))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 51, 100, 'c2', 1, 50))
        self.g.add_edge(edge.Edge('c2', 41, 80, 'c3', 1, 40))
        self.assertTrue(self.g._remove_middle_node('c1', 'c2', 'c3'))
        self.assertTrue('c2' not in self.g.graph)
        self.assertEqual(1, len(self.g.graph['c1']['c3']['edges']))
        self.assertEqual(1, len(self.g.graph['c3']['c1']['edges']))
        self.assertEqual(self.g.graph['c1']['c3']['edges'][0], edge.Edge('c1', 91, 100, 'c3', 1, 10))
        self.assertEqual(self.g.graph['c3']['c1']['edges'][0], edge.Edge('c1', 91, 100, 'c3', 1, 10))


    def test_remove_redundant_nodes_from_simple_path(self):
        '''test remove_redundant_nodes_from_simple_path'''
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 100, 142, 'c3', 1, 42))
        new_path = self.g.remove_redundant_nodes_from_simple_path(['c1', 'c2', 'c3'])
        self.assertListEqual(['c1', 'c2', 'c3'], new_path)
        self.assertTrue('c2' in self.g.graph)

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 51, 100, 'c2', 1, 50))
        self.g.add_edge(edge.Edge('c2', 41, 80, 'c3', 1, 40))
        new_path = self.g.remove_redundant_nodes_from_simple_path(['c1', 'c2', 'c3'])
        self.assertListEqual(['c1', 'c3'], new_path)
        self.assertFalse('c2' in self.g.graph)

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 51, 100, 'c2', 1, 50))
        self.g.add_edge(edge.Edge('c2', 41, 80, 'c3', 1, 40))
        self.g.add_edge(edge.Edge('c3', 100, 150, 'c4', 1, 50))
        self.g.add_edge(edge.Edge('c4', 20, 70, 'c5', 1, 50))
        self.g.add_edge(edge.Edge('c5', 40, 90, 'c6', 1, 50))
        new_path = self.g.remove_redundant_nodes_from_simple_path(['c1', 'c2', 'c3', 'c4', 'c5', 'c6'])
        self.assertListEqual(['c1', 'c3', 'c5', 'c6'], new_path)
        self.assertFalse('c2' in self.g.graph)
        self.assertFalse('c4' in self.g.graph)


    def test_node_to_coords(self):
        '''test _node_to_coords'''
        self.g.add_edge(edge.Edge('c1', 199, 0, 'c2', 1319, 1119))
        nodes = ['c1', 'c2']
        self.assertListEqual(['c1', intervals.Interval(200, 659), True], self.g._node_to_coords(nodes, 0))
        self.assertListEqual(['c2', intervals.Interval(0, 1319), True], self.g._node_to_coords(nodes, 1))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c2', 1119, 1319, 'c1', 0, 199))
        nodes = ['c1', 'c2']
        self.assertListEqual(['c1', intervals.Interval(200, 659), True], self.g._node_to_coords(nodes, 0))
        self.assertListEqual(['c2', intervals.Interval(0, 1319), True], self.g._node_to_coords(nodes, 1))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 551, 650, 'c2', 1, 100))
        self.g.add_edge(edge.Edge('c2', 1201, 1300, 'c3', 11, 110))
        nodes = ['c1', 'c2', 'c3']
        self.assertListEqual(['c1', intervals.Interval(0, 550), False], self.g._node_to_coords(nodes, 0))
        self.assertListEqual(['c2', intervals.Interval(1, 1200), False], self.g._node_to_coords(nodes, 1))
        self.assertListEqual(['c3', intervals.Interval(11, 2159), False], self.g._node_to_coords(nodes, 2))


        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 551, 650, 'c2', 1, 100))
        self.g.add_edge(edge.Edge('c3', 110, 11, 'c2', 1300, 1201))
        self.g.add_edge(edge.Edge('c3', 2051, 2150, 'c4', 6, 115))
        nodes = ['c1', 'c2', 'c3', 'c4']
        self.assertListEqual(['c1', intervals.Interval(0, 550), False], self.g._node_to_coords(nodes, 0))
        self.assertListEqual(['c2', intervals.Interval(1, 1200), False], self.g._node_to_coords(nodes, 1))
        self.assertListEqual(['c3', intervals.Interval(11, 2050), False], self.g._node_to_coords(nodes, 2))
        self.assertListEqual(['c4', intervals.Interval(6, 1259), False], self.g._node_to_coords(nodes, 3))


    def test_merged_coords_from_simple_nonredundant_path(self):
        '''test merged_coords_from_simple_nonredundant_path'''
        self.g.add_edge(edge.Edge('c1', 199, 0, 'c2', 1319, 1119))
        nodes = ['c1', 'c2']
        coords = self.g.merged_coords_from_simple_nonredundant_path(nodes)
        expected = [
            ['c1', intervals.Interval(200, 659), True],
            ['c2', intervals.Interval(0, 1319), True],
        ]
        self.assertListEqual(expected, coords)


        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 610, 652, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 1250, 1310, 'c3', 5, 65))
        nodes = ['c1', 'c2', 'c3']
        coords = self.g.merged_coords_from_simple_nonredundant_path(nodes)
        expected = [
            ['c1', intervals.Interval(0, 609), False],
            ['c2', intervals.Interval(1, 1249), False],
            ['c3', intervals.Interval(5, 2159), False]
        ]
        self.assertListEqual(expected, coords)


    def test_simple_path_is_consistent(self):
        '''test simple_path_is_consistent'''
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertTrue(self.g.simple_path_is_consistent(['c1', 'c2']))
        self.g.add_edge(edge.Edge('c3', 10, 100, 'c4', 1, 1000))
        self.assertTrue(self.g.simple_path_is_consistent(['c1', 'c2', 'c3']))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 1000, 100, 'c3', 1, 42))
        self.assertFalse(self.g.simple_path_is_consistent(['c1', 'c2', 'c3']))


    def test_edges_are_consistent(self):
        '''test edges_are_consistent'''
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertTrue(self.g._edges_are_consistent('c1', 'c2', 'c3'))
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c3', 1, 42))
        self.assertFalse(self.g._edges_are_consistent('c1', 'c2', 'c3'))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertFalse(self.g._edges_are_consistent('c1', 'c2', 'c3'))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertFalse(self.g._edges_are_consistent('c1', 'c2', 'c3'))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 1, 42))
        self.g.add_edge(edge.Edge('c2', 100, 50, 'c3', 1, 42))
        self.assertFalse(self.g._edges_are_consistent('c1', 'c2', 'c3'))

        self.g = graph.Graph(self.asm)
        self.g.add_edge(edge.Edge('c1', 1, 100, 'c2', 42, 1))
        self.g.add_edge(edge.Edge('c2', 50, 100, 'c3', 1, 42))
        self.assertFalse(self.g._edges_are_consistent('c1', 'c2', 'c3'))




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
__all__ = [
    'assembly',
    'common',
    'contig',
    'contig_trim',
    'egg_extract',
    'edge',
    'external_progs',
    'graph',
    'kcount',
    'kmers',
    'kraken',
    'mapping',
    'mummer',
    'qc',
    'qc_external',
    'read_trim',
    'seed',
    'seed_processor',
    'test_data_runner',
]
from iva import *



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
#!/usr/bin/env python3

import unittest
import os
from iva import contig
from pyfastaq import sequences

class TestContig(unittest.TestCase):
    def test_init(self):
        '''Test init'''
        ctg = contig.Contig(sequences.Fasta('ID', 'ACCGT'))
        self.assertTrue(ctg.fa, sequences.Fasta('ID', 'ACCGT'))


    def test_len(self):
        '''Test len'''
        ctg = contig.Contig(sequences.Fasta('ID', 'ACCGT'))
        self.assertEqual(len(ctg), 5)


    def test_add_left_kmer(self):
        '''Test add_left_kmer'''
        ctg = contig.Contig(sequences.Fasta('ID', 'ACCGT'))
        ctg.add_left_kmer('GT')
        self.assertListEqual(ctg.left_kmers.kmers, ['TG'])


    def test_add_right_kmer(self):
        '''Test add_right_kmer'''
        ctg = contig.Contig(sequences.Fasta('ID', 'ACCGT'))
        ctg.add_right_kmer('GT')
        self.assertListEqual(ctg.right_kmers.kmers, ['GT'])


    def test_extend(self):
        '''Test extend'''
        ctg = contig.Contig(sequences.Fasta('ID', 'ACCGT'))
        self.assertEqual(ctg.extend(5, 2, 100), (0, 0))
        self.assertEqual(ctg.fa, sequences.Fasta('ID', 'ACCGT'))
        ctg.add_left_kmer('GT')
        self.assertEqual(ctg.extend(1, 2, 100), (2, 0))
        self.assertEqual(ctg.fa, sequences.Fasta('ID', 'GTACCGT'))
        self.assertEqual(ctg.extend(1, 2, 100), (0, 0))
        self.assertEqual(ctg.fa, sequences.Fasta('ID', 'GTACCGT'))
        ctg.add_right_kmer('TG')
        self.assertEqual(ctg.extend(1, 2, 100), (0, 2))
        self.assertEqual(ctg.fa, sequences.Fasta('ID', 'GTACCGTTG'))
        self.assertEqual(ctg.extend(1, 2, 100), (0, 0))
        self.assertEqual(ctg.fa, sequences.Fasta('ID', 'GTACCGTTG'))
        ctg.add_left_kmer('AG')
        ctg.add_right_kmer('GC')
        self.assertEqual(ctg.extend(1, 2, 100), (2, 2))
        self.assertEqual(ctg.fa, sequences.Fasta('ID', 'AGGTACCGTTGGC'))




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import os
from iva import common

def run_trimmomatic(reads1, reads2, outprefix, trimmo_jar, adapters, minlen=50, verbose=0, threads=1, qual_trim=''):
    cmd = ' '.join([
        'java -Xmx1000m -jar',
        trimmo_jar,
        'PE',
        '-threads', str(threads),
        reads1,
        reads2,
        outprefix + '_1.fq',
        outprefix + '.unpaired_1.fq',
        outprefix + '_2.fq',
        outprefix + '.unpaired_2.fq',
        'ILLUMINACLIP:' + os.path.abspath(adapters) + ':2:10:7:1',
        qual_trim,
        'MINLEN:' + str(minlen)
    ])

    if verbose:
        print('Run trimmomatic:', cmd)
    common.syscall(cmd)
    os.unlink(outprefix + '.unpaired_1.fq')
    os.unlink(outprefix + '.unpaired_2.fq')



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import filecmp
from iva import kcount

modules_dir = os.path.dirname(os.path.abspath(kcount.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
kcount.verbose = 2

class TestKcount(unittest.TestCase):
    def test_head_fastaq_fq(self):
        '''Test head_fastaq with FASTQ input'''
        test_out = 'tmp.head.fa'
        reads1 = os.path.join(data_dir, 'kcount_test.reads_1.fastq')
        reads2 = os.path.join(data_dir, 'kcount_test.reads_2.fastq')
        kcount._head_fastaq(reads1, reads2, test_out, count=4)
        self.assertTrue(filecmp.cmp(test_out, os.path.join(data_dir, 'kcount_test.reads.head.4.fa'), shallow=False))
        os.unlink(test_out)


    def test_head_fastaq_fa(self):
        '''Test head_fastaq with FASTA input'''
        test_out = 'tmp.head.fa'
        reads1 = os.path.join(data_dir, 'kcount_test.reads_1.fasta')
        reads2 = os.path.join(data_dir, 'kcount_test.reads_2.fasta')
        kcount._head_fastaq(reads1, reads2, test_out, count=4)
        self.assertTrue(filecmp.cmp(test_out, os.path.join(data_dir, 'kcount_test.reads.head.4.fa'), shallow=False))
        os.unlink(test_out)


    def test_median(self):
        '''Test _median()'''
        d = {1: 1, 2: 2, 3: 5, 4: 2, 5: 1}
        self.assertEqual(3, kcount._median(d))


    def test_run_kmc(self):
        '''Test test_run_kmc'''
        reads = os.path.join(data_dir, 'kcount_test.run_kmc.fa')
        counts_file = kcount._run_kmc(reads, 'tmp.run_kmc', 10, 2, 4)
        self.assertTrue(filecmp.cmp(counts_file, os.path.join(data_dir, 'kcount_test.run_kmc.counts'), shallow=False))
        os.unlink(counts_file)


    def test_run_kmc_two_threads(self):
        '''Test test_run_kmc with two threads'''
        reads = os.path.join(data_dir, 'kcount_test.run_kmc.fa')
        counts_file = kcount._run_kmc(reads, 'tmp.run_kmc', 10, 2, 4, threads=2)
        self.assertTrue(filecmp.cmp(counts_file, os.path.join(data_dir, 'kcount_test.run_kmc.counts'), shallow=False))
        os.unlink(counts_file)


    def test_kmc_to_kmer_counts(self):
        '''Test _kmc_to_kmer_counts'''
        counts = kcount._kmc_to_kmer_counts(os.path.join(data_dir, 'kcount_test.kmc_counts'), number=2)
        expected = {'ACGT': 10, 'ATGC': 10}
        self.assertDictEqual(counts, expected)


    def test_counts_file_to_fasta(self):
        '''Test _counts_file_to_fasta'''
        outfile = 'tmp.kmer_counts_to_fa.fa'
        expected = os.path.join(data_dir, 'kcount_test.kmc_counts.fa')
        infile = os.path.join(data_dir, 'kcount_test.kmc_counts')
        kcount._counts_file_to_fasta(infile, outfile)
        self.assertTrue(filecmp.cmp(outfile, expected, shallow=False))
        os.unlink(outfile)


    def test_get_most_common_kmers(self):
        '''Test get_most_common_kmers'''
        reads1 = os.path.join(data_dir, 'kcount_test.get_commonest_kmer_1.fa')
        reads2 = os.path.join(data_dir, 'kcount_test.get_commonest_kmer_2.fa')
        counts = kcount.get_most_common_kmers(reads1, reads2, kmer_length=10, head=100000, min_count=2, max_count=4, most_common=100, method='kmc')
        self.assertDictEqual({'AGCTAAAACT': 2, 'CTATATCTCA': 3}, counts)




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import filecmp
import shutil
import pyfastaq
from iva import seed

modules_dir = os.path.dirname(os.path.abspath(seed.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestSeed(unittest.TestCase):
    def test_kmc_in_path(self):
        '''Test that kmc is in the user's path'''
        assert(shutil.which('kmc') is not None)
        assert(shutil.which('kmc_dump') is not None)


    def test_init(self):
        '''test init'''
        # TODO


    def test_extension_from_read(self):
        '''Test _test_extension_from_read'''
        s = seed.Seed(seq='AGGCT')
        self.assertEqual(None, s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AAAAA')))
        self.assertEqual(None, s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AGGC')))
        self.assertEqual('A', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AGGCTA')))
        self.assertEqual('AT', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AGGCTAT')))
        self.assertEqual('AT', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'GGGAGGCTAT')))
        self.assertEqual('AA', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'TTAGCCT')))

        self.assertEqual(None, s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AAAAA'), left=True))
        self.assertEqual(None, s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AGGCTA'), left=True))
        self.assertEqual('GT', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'GTAGGCTA'), left=True))
        self.assertEqual('GT', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'GTAGGCTATTC'), left=True))
        self.assertEqual('GT', s._extension_from_read(pyfastaq.sequences.Fasta('x', 'AGCCTAC'), left=True))


    def test_len(self):
        '''Test len'''
        s = seed.Seed(seq='AGGCT')
        self.assertEqual(5, len(s))
        s.seq = None
        self.assertEqual(0, len(s))


    def test_extensions_from_reads_file(self):
        '''Test _extensions_from_reads_file'''
        s = seed.Seed(seq='AGGCT')
        l, r = s._extensions_from_reads_file(os.path.join(data_dir, 'kcount_test.reads_1.fasta'))
        self.assertListEqual(l, [])
        self.assertListEqual(r, ['A', 'AT', 'AT'])
        l, r = s._extensions_from_reads_file(os.path.join(data_dir, 'kcount_test.reads_2.fasta'))
        self.assertListEqual(l, ['G', 'TG', 'TG'])
        self.assertListEqual(r, [])


    def test_extend_with_reads_as_single_end(self):
        '''Test _extend_with_reads_as_single_end'''
        s = seed.Seed(seq='AGGCT', ext_min_cov=1, verbose=2)
        reads1 = os.path.join(data_dir, 'kcount_test.reads_1.fasta')
        reads2 = os.path.join(data_dir, 'kcount_test.reads_2.fasta')
        s._extend_with_reads_as_single_end(reads1, reads2)
        self.assertEqual('TGAGGCTAT', s.seq)


    def test_extend(self):
        '''Test extend'''
        s = seed.Seed(seq='AGGCT', ext_min_cov=1, verbose=3)
        reads1 = os.path.join(data_dir, 'kcount_test.reads_1.fasta')
        reads2 = os.path.join(data_dir, 'kcount_test.reads_2.fasta')
        s.extend(reads1, reads2, 100)
        self.assertEqual('TGAGGCTAT', s.seq)


    def test_write_fasta(self):
        '''Test write_fasta'''
        s = seed.Seed(seq='GAAGGCGGCAGC')
        tmpfile = 'tmp.seed.fa'
        s.write_fasta(tmpfile, 'spam')
        self.assertTrue(filecmp.cmp(tmpfile, os.path.join(data_dir, 'seed_test.write_fasta.fa'), shallow=False))
        os.unlink(tmpfile)



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import os
import unittest
from iva import read_trim

modules_dir = os.path.dirname(os.path.abspath(read_trim.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestReadTrim(unittest.TestCase):
    def test_run_trimmomatic(self):
        '''Test run_trimmomatic'''
        reads1 = os.path.join(data_dir, 'read_trim_test.reads_1.fq')
        reads2 = os.path.join(data_dir, 'read_trim_test.reads_2.fq')
        #run_trimmomatic(reads1, reads2, outprefix, trimmo_jar, adapters, minlen=50, verbose=0):
        # need to know where trimmoatic jar file is - could be anywhere - and
        # read trimming is optional, so skip this test for now...
        pass



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import os
import iva

class Error (Exception): pass

class Tester:
    def __init__(self, outdir, iva_script, trimmo_jar=None, threads=1):
        self.outdir = os.path.join(outdir)
        if os.path.exists(self.outdir):
            raise Error('Output directory already exists. Cannot continue')

        self.iva_script = iva_script
        self.trimmo_jar = trimmo_jar
        self.threads = threads


    def _copy_input_files(self):
        extractor = iva.egg_extract.Extractor(os.path.abspath(os.path.join(os.path.dirname(iva.__file__), os.pardir)))
        test_files = os.path.join('iva', 'test_run_data')
        extractor.copy_dir(test_files, self.outdir)
        print('Copied input test files into here:', os.path.abspath(self.outdir))
         

    def _run_iva(self):
        os.chdir(self.outdir)
        cmd = self.iva_script + ' --threads ' + str(self.threads)
        if self.trimmo_jar:
            cmd += ' --trimmomatic ' + self.trimmo_jar

        cmd += ' --pcr_primers hiv_pcr_primers.fa -f reads_1.fq.gz -r reads_2.fq.gz iva.out'

        print('Current working directory:', os.getcwd())
        print('Running iva on the test data with the command:', cmd, sep='\n')
        iva.common.syscall(cmd)


    def _check_output(self):
        print('Finished running iva')
        expected_contigs_file = os.path.abspath(os.path.join('iva.out', 'contigs.fasta'))
        if os.path.exists(expected_contigs_file):
            print('Looks OK. Final output contigs file is:', expected_contigs_file)
        else:
            print('Something went wrong! Final output contigs file not found:', expected_contigs_file)


    def run(self):
        self._copy_input_files()
        self._run_iva()
        self._check_output()




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import sys
import os
import filecmp
import pysam
from iva import assembly

modules_dir = os.path.dirname(os.path.abspath(assembly.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestAssembly(unittest.TestCase):
    def test_init_and_write(self):
        '''test assembly initialise and write to file'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        tmp_out = 'tmp.assembly_test.out.fa'
        a.write_contigs_to_file(tmp_out)
        self.assertTrue(filecmp.cmp(tmp_out, os.path.join(data_dir, 'assembly_test.fa')))
        os.unlink(tmp_out)


    def test_map_reads(self):
        '''test _map_reads'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        reads_prefix = os.path.join(data_dir, 'assembly_test.to_map')
        out_prefix = 'tmp.assembly_test.out'
        a._map_reads(reads_prefix + '_1.fastq', reads_prefix + '_2.fastq', out_prefix)

        # different smalt version output slightly different BAMs. Some columns
        # should never change, so check just those ones
        def get_sam_columns(bamfile):
            sams = []
            sam_reader = pysam.Samfile(bamfile, "rb")
            for sam in sam_reader.fetch(until_eof=True):
                if sam.is_unmapped:
                    refname = None
                else:
                    refname = sam_reader.getrname(sam.tid)
                sams.append((sam.qname, sam.flag, refname, sam.pos, sam.cigar, sam.seq))
            return sams

        expected = get_sam_columns(os.path.join(data_dir, 'assembly_test.mapped.bam'))
        got = get_sam_columns(out_prefix + '.bam')
        self.assertListEqual(expected, got)
        os.unlink(out_prefix + '.bam')


    def test_extend_contigs_with_bam(self):
        '''test _extend_contigs_with_bam'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'mapping_test.ref.trimmed.fa'), ext_min_cov=1, ext_min_ratio=1, ext_bases=10, max_insert=200)
        bam = os.path.join(data_dir, 'mapping_test.smalt.out.bam')
        out_prefix = 'tmp'
        a._extend_contigs_with_bam(bam, out_prefix, output_all_useful_reads=False)
        tmp_contigs = 'tmp.new_contigs.fa'
        tmp_reads_1 = out_prefix + '_1.fa'
        tmp_reads_2 = out_prefix + '_2.fa'
        a.write_contigs_to_file(tmp_contigs)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'mapping_test.ref.fa'), tmp_contigs))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'assembly_test.extend_kept_reads_1.fa'), tmp_reads_1))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'assembly_test.extend_kept_reads_2.fa'), tmp_reads_2))
        os.unlink(tmp_contigs)
        os.unlink(tmp_reads_1)
        os.unlink(tmp_reads_2)


    def test_contig_worth_extending(self):
        '''Test _contig_worth_extending'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        self.assertTrue(a._contig_worth_extending('A'))
        self.assertTrue(a._contig_worth_extending('B'))
        self.assertTrue(a._contig_worth_extending('C'))
        a.contig_lengths['A'].append([100, 0, 0])
        self.assertTrue(a._contig_worth_extending('A'))
        a.contig_lengths['A'].append([100, 0, 0])
        self.assertFalse(a._contig_worth_extending('A'))
        a.contig_lengths['A'].append([100, 0, 0])
        self.assertFalse(a._contig_worth_extending('A'))
        a.contig_lengths['A'][-1] = [101, 1, 0]
        self.assertTrue(a._contig_worth_extending('A'))


    def test_worth_extending(self):
        '''Test worth_extending'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        self.assertTrue(a._worth_extending())
        for x in ['A', 'B', 'C']:
            a.contig_lengths[x].append([100, 0, 0])
        self.assertTrue(a._worth_extending())
        for x in ['A', 'B', 'C']:
            a.contig_lengths[x].append([100, 0, 0])
        self.assertFalse(a._worth_extending())
        for x in ['A', 'B']:
            a.contig_lengths[x].append([100, 0, 0])
        self.assertFalse(a._worth_extending())
        a.contig_lengths['C'].append([100, 0, 0])
        self.assertFalse(a._worth_extending())


    def test_read_pair_extend(self):
        '''Test read_pair_extend'''
        ref = os.path.join(data_dir, 'assembly_test_read_pair_extend.ref.fasta')
        to_extend = os.path.join(data_dir, 'assembly_test_read_pair_extend.to_extend.fasta')
        reads_prefix = os.path.join(data_dir, 'assembly_test_read_pair_extend.ref.reads')
        a = assembly.Assembly(contigs_file=to_extend, strand_bias=0, verbose=4)
        a.read_pair_extend(reads_prefix, 'tmp.extend')
        self.assertTrue(len(a.contigs['1']) > 900)


    def test_get_ref_length(self):
        '''Test _get_ref_length'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        sam_reader = pysam.Samfile(os.path.join(data_dir, 'assembly_test.mapped.bam'), "rb")
        expected_lengths = [100] * 6 + [None, None]
        i = 0
        for sam in sam_reader.fetch(until_eof=True):
            self.assertEqual(expected_lengths[i], a._get_ref_length(sam_reader, sam))
            i += 1


    def test_get_ref_length_sam_pair(self):
        '''Test _get_ref_length_sam_pair'''
        previous_sam = None
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        sam_reader = pysam.Samfile(os.path.join(data_dir, 'assembly_test.mapped.bam'), "rb")
        expected_lengths = [100] * 3 + [None]
        i = 0
        for current_sam in sam_reader.fetch(until_eof=True):
            if previous_sam is None:
                previous_sam = current_sam
                continue
            self.assertEqual(expected_lengths[i], a._get_ref_length_sam_pair(sam_reader, current_sam, previous_sam))
            i += 1
            previous_sam = None


    def test_get_unmapped_pairs(self):
        '''Test _get_unmapped_pairs'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'))
        reads1 = os.path.join(data_dir, 'assembly_test.to_map_1.fastq')
        reads2 = os.path.join(data_dir, 'assembly_test.to_map_2.fastq')
        a._get_unmapped_pairs(reads1, reads2, 'tmp')
        self.assertTrue(filecmp.cmp('tmp_1.fa', os.path.join(data_dir, 'assembly_test_get_unmapped_pairs_1.fa'), shallow=False))
        self.assertTrue(filecmp.cmp('tmp_2.fa', os.path.join(data_dir, 'assembly_test_get_unmapped_pairs_2.fa'), shallow=False))
        os.unlink('tmp_1.fa')
        os.unlink('tmp_2.fa')


    def test_add_new_seed_contig(self):
        '''Test add_new_seed_contig'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test.fa'), verbose=3, min_clip=1, seed_min_cov=1, seed_min_kmer_count=1, seed_start_length=10, seed_overlap_length=5, seed_stop_length=20)
        reads1 = os.path.join(data_dir, 'assembly_test_add_new_seed_contig.reads_1.fa')
        reads2 = os.path.join(data_dir, 'assembly_test_add_new_seed_contig.reads_2.fa')
        new_contig_name = a.add_new_seed_contig(reads1, reads2)
        self.assertEqual('seeded.00001', new_contig_name)
        self.assertTrue(new_contig_name in a.contigs)


    def test_good_intervals_from_strand_coverage(self):
        '''Test good_intervals_from_strand_coverage'''
        fwd_cov = [0, 1, 1, 2, 5, 10, 100, 10, 10, 6, 0, 10, 10, 10, 5, 10]
        rev_cov = [0, 5, 5, 5, 5, 20, 10,  10, 10, 100, 9, 10, 10, 10, 5, 0]
        expected = [(3,5), (7,8), (11,14)]
        a = assembly.Assembly(strand_bias=0.2)
        got = a._good_intervals_from_strand_coverage(fwd_cov, rev_cov)
        self.assertListEqual(expected, got)


    def test_get_contig_order_by_orfs(self):
        '''Test get_contig_order_by_orfs'''
        a = assembly.Assembly(contigs_file=os.path.join(data_dir, 'assembly_test_order_by_orfs.fa'))
        got = a._get_contig_order_by_orfs(min_length=240)
        expected = [('1', True), ('3', False), ('4', False), ('6', False), ('2', False), ('5', False)]
        self.assertListEqual(expected, got)


    def test_trim_contig_for_strand_bias(self):
        '''Test _trim_contig_for_strand_bias'''
        # TODO
        pass


    def test_subcontigs_from_strand_bias(self):
        '''Test _subcontigs_from_strand_bias'''
        # TODO
        pass


    def test_trim_strand_biased_ends(self):
        '''Test _trim_strand_biased_ends'''
        # TODO
        pass


    def test_trim_contigs(self):
        '''Test trim_contigs'''
        # TODO
        pass


    def test_remove_contained_contigs(self):
        '''Test _remove_contained_contigs'''
        # TODO
        pass


    def test_coords_to_new_contig(self):
        '''Test _coords_to_new_contig'''
        # TODO
        pass


    def test_merge_overlapping_contigs(self):
        '''Test _merge_overlapping_contigs'''
        # TODO
        pass


    def test_contig_names_size_order(self):
        '''Test _contig_names_size_order'''
        # TODO
        pass


    def test_contig_contained_in_nucmer_hits(self):
        '''Test _contig_contained_in_nucmer_hits'''
        # TODO
        pass


    def test_remove_contig_from_nucmer_hits(self):
        '''Test _remove_contig_from_nucmer_hits'''
        # TODO
        pass


    def test_remove_contig(self):
        '''Test _remove_contig'''
        # TODO
        pass



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
from iva import kmers

class TestKmers(unittest.TestCase):
    def test_init(self):
        '''Test __init__'''
        k = kmers.Kmers(kmer='ACGGT')
        self.assertListEqual(['ACGGT'], k.kmers)
        k = kmers.Kmers(kmer='ACGGT', left=True)
        self.assertListEqual(['TGGCA'], k.kmers)


    def test_append(self):
        '''Test append'''
        k = kmers.Kmers(kmer='A')
        k.append('G')
        self.assertListEqual(['A', 'G'], k.kmers)
        k.append('N')
        self.assertListEqual(['A', 'G'], k.kmers)


    def test_extend(self):
        '''Test extend'''
        k = kmers.Kmers(kmer='A')
        k.extend(['A', 'C'])
        self.assertListEqual(['A', 'A', 'C'], k.kmers)


    def test_kmer_dict(self):
        '''Test kmer_dict'''
        k = kmers.Kmers(kmer='A')
        k.extend(['A', 'C', 'GT', 'GC', 'ACT'])
        expected = [
            {'A': 3, 'C': 1, 'G': 2},
            {'GT': 1, 'GC': 1, 'AC': 1},
            {'ACT': 1}
        ]

        for i in range(3):
            d = k._kmer_dict(i+1)
            self.assertDictEqual(d, expected[i])


    def test_commonest_kmers(self):
        k = kmers.Kmers(kmer='AC')
        k.extend(['AT'] * 4)
        k.append('ATT')

        commonest_kmers, counts = k._commonest_kmers(1)
        self.assertEqual(commonest_kmers, (None, 'A'))
        self.assertEqual(counts, (None, 6))

        commonest_kmers, counts = k._commonest_kmers(2)
        self.assertEqual(commonest_kmers, ('AC', 'AT'))
        self.assertEqual(counts, (1, 5))

        commonest_kmers, counts = k._commonest_kmers(3)
        self.assertEqual(commonest_kmers, (None, 'ATT'))
        self.assertEqual(counts, (None, 1))

        commonest_kmers, counts = k._commonest_kmers(4)
        self.assertEqual(commonest_kmers, (None, None))
        self.assertEqual(counts, (None, None))


    def test_extension(self):
        '''Test extension'''
        k = kmers.Kmers(kmer='AT', left=True)
        self.assertEqual(k.extension(1, 2, 100), 'AT')
        k.extend(['GGT'] * 5)
        self.assertEqual(k.extension(5, 2, 100), 'GGT')
        self.assertEqual(k.extension(5, 2, 2), 'GT')

        k = kmers.Kmers()
        self.assertEqual(k.extension(5, 2, 100), '')
        k.extend(['ACG'] * 5)
        self.assertEqual(k.extension(5, 2, 100), 'ACG')
        self.assertEqual(k.extension(6, 2, 100), '')
        self.assertEqual(k.extension(5, 2, 2), 'AC')
        self.assertEqual(k.extension(5, 2, 1), 'A')

        k.extend(['ACC'] * 2)
        self.assertEqual(k.extension(5, 2, 100), 'ACG')
        self.assertEqual(k.extension(5, 3, 100), 'AC')




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import sys
import shutil
import os
import filecmp
from iva import egg_extract

modules_dir = os.path.dirname(os.path.abspath(egg_extract.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestEggExtract(unittest.TestCase):
    def setUp(self):
        self.zipped = os.path.join(data_dir, 'egg_extract_test_dir.zip')
        self.not_zipped = os.path.join(data_dir, 'egg_extract_test_dir')
        self.egg_zipped = egg_extract.Extractor(self.zipped)
        self.egg_not_zipped = egg_extract.Extractor(self.not_zipped)


    def test_init_and_write(self):
        '''test __init__'''
        with self.assertRaises(egg_extract.Error):
            egg_extract.Extractor('notafilethrowerror')

        self.assertTrue(self.egg_zipped.zip_file is not None)
        self.assertTrue(self.egg_not_zipped.zip_file is None)


    def test_copy_file_unzipped(self):
        '''test _copy_file_unzipped'''
        to_copy = os.path.join('iva', 'gage', 'gage1')
        outfile = 'tmp.copy_file_unzipped'
        if os.path.exists(outfile):
            os.unlink(outfile)
        self.egg_not_zipped._copy_file_unzipped(to_copy, outfile)
        self.assertTrue(os.path.exists(outfile))
        os.unlink(outfile)


    def test_copy_file_zipped(self):
        '''test _copy_file_zipped'''
        outfile = 'tmp.copy_file_zipped'
        if os.path.exists(outfile):
            os.unlink(outfile)
        to_copy = os.path.join('iva', 'gage', 'gage1')
        self.egg_zipped._copy_file_zipped(to_copy, outfile)
        self.assertTrue(os.path.exists(outfile))
        os.unlink(outfile)


    def test_copy_file(self):
        '''test copy_file'''
        outfile = 'tmp.copy_file'
        for egg in [self.egg_zipped, self.egg_not_zipped]:
            if os.path.exists(outfile):
                os.unlink(outfile)
            to_copy = os.path.join('iva', 'gage', 'gage1')
            egg.copy_file(to_copy, outfile)
            self.assertTrue(os.path.exists(outfile))
            os.unlink(outfile)


    def test_copy_dir_unzipped(self):
        '''test copy_dir_unzipped'''
        outdir = 'tmp.copy_dir_unzipped'
        to_copy = os.path.join('iva', 'gage')
        if os.path.exists(outdir):
            shutil.rmtree(outdir)

        self.egg_not_zipped._copy_dir_unzipped(to_copy, outdir)
        self.assertTrue(os.path.exists(outdir))
        self.assertTrue(os.path.isdir(outdir))
        self.assertTrue(os.path.exists(os.path.join(outdir, 'gage1')))
        self.assertTrue(os.path.exists(os.path.join(outdir, 'gage2')))
        shutil.rmtree(outdir)


    def test_copy_dir_zipped(self):
        '''test copy_dir_zipped'''
        outdir = 'tmp.copy_dir_zipped'
        to_copy = os.path.join('iva', 'gage')
        if os.path.exists(outdir):
            shutil.rmtree(outdir)

        self.egg_zipped._copy_dir_zipped(to_copy, outdir)
        self.assertTrue(os.path.exists(outdir))
        self.assertTrue(os.path.isdir(outdir))
        self.assertTrue(os.path.exists(os.path.join(outdir, 'gage1')))
        self.assertTrue(os.path.exists(os.path.join(outdir, 'gage2')))
        shutil.rmtree(outdir)


    def test_copy_dir(self):
        '''test copy_dir'''
        outdir = 'tmp.copy_dir'
        for egg in [self.egg_zipped, self.egg_not_zipped]:
            if os.path.exists(outdir):
                shutil.rmtree(outdir)
            to_copy = os.path.join('iva', 'gage')
            egg.copy_dir(to_copy, outdir)
            self.assertTrue(os.path.exists(outdir))
            self.assertTrue(os.path.isdir(outdir))
            self.assertTrue(os.path.exists(os.path.join(outdir, 'gage1')))
            self.assertTrue(os.path.exists(os.path.join(outdir, 'gage2')))
            shutil.rmtree(outdir)




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import filecmp
import pyfastaq
from iva import seed_processor

modules_dir = os.path.dirname(os.path.abspath(seed_processor.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestSeedProcessor(unittest.TestCase):
    def test_process_seeds(self):
        '''Test process_seeds'''
        in_fasta = os.path.join(data_dir, 'seed_processor_test.ref.fa')
        reads1 = os.path.join(data_dir, 'seed_processor_test.reads_1.fa')
        reads2 = os.path.join(data_dir, 'seed_processor_test.reads_2.fa')
        tmp_out = 'tmp.process_seeds.out'
        s = seed_processor.SeedProcessor(in_fasta, reads1, reads2, tmp_out, verbose=4, threads=1)
        s.process()
        reader = pyfastaq.sequences.file_reader(tmp_out)
        counter = 0
        for seq in reader:
            self.assertTrue(len(seq) > 470)
            counter += 1
        self.assertEqual(2, counter)
        os.unlink(tmp_out)



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import filecmp
import os
import copy
from iva import contig_trim

modules_dir = os.path.dirname(os.path.abspath(contig_trim.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')



class TestContigTrim(unittest.TestCase):
    def test_coverage_to_trimmed_coords(self):
        '''Test _coverage_to_trimmed_coords'''
        coverage = [
            [0] * 4,
            [0] * 5,
            [0] * 6,
            [0] * 20,
            [1] * 5 + [0] * 15,
            [0] * 15 + [1] * 5,
            [0] * 2 + [1] * 5 + [0] * 13,
            [0] * 3 + [1] * 5 + [0] * 12,
            [0] * 4 + [1] * 5 + [0] * 11,
            [0] * 5 + [1] * 5 + [0] * 10,
            [1] * 20,
            [0] * 2 + [1] * 16 + [0] * 2,
        ]

        expected = [
            (0, 3),
            (0, 4),
            (0, 5),
            (0, 19),
            (5, 19),
            (0, 14),
            (7, 19),
            (8, 19),
            (9, 19),
            (0, 19),
            None,
            None,
        ]

        assert len(coverage) == len(expected)

        for i in range(len(coverage)):
            got = contig_trim._coverage_to_trimmed_coords(coverage[i], min_dist_to_end=3, window_length=5, min_pc=80)

            # Should be able to reverse the coverage and still get the same results (but also reversed)
            coverage_rev = copy.copy(coverage[i])
            coverage_rev.reverse()
            if expected[i] is None:
                expected_rev = None
            else:
                expected_rev = [len(coverage[i]) - x - 1 for x in expected[i]]
                expected_rev.reverse()
                expected_rev = tuple(expected_rev)
        got_rev = contig_trim._coverage_to_trimmed_coords(coverage_rev, min_dist_to_end=3, window_length=5, min_pc=80)
        

    def test_trim_ends(self):
        '''Test _trim_ends'''
        before_trim = os.path.join(data_dir, 'contig_trim_test_contigs.fa')
        expected_after_trim = os.path.join(data_dir, 'contig_trim_test_contigs.trimmed.fa')
        adapters = os.path.join(data_dir, 'contig_trim_test_contigs.adapters_and_primers.fa')
        tmp_out = 'tmp.trimmed.fa'
        contig_trim._trim_ends(before_trim, tmp_out, adapters, min_length=20, min_dist_to_end=5, window_length=10, min_pc=90)
        self.assertTrue(filecmp.cmp(tmp_out, expected_after_trim, shallow=False))
        os.unlink(tmp_out)


    def test_trim_primers_and_adapters(self):
        '''Test trim_primers_and_adapters'''
        before_trim = os.path.join(data_dir, 'contig_trim_test_contigs.fa')
        expected_after_trim = os.path.join(data_dir, 'contig_trim_test_contigs.trimmed.fa')
        adapters = os.path.join(data_dir, 'contig_trim_test_contigs.adapters.fa')
        primers = os.path.join(data_dir, 'contig_trim_test_contigs.primers.fa')
        tmp_out = 'tmp.trimmed.fa'
        contig_trim.trim_primers_and_adapters(before_trim, tmp_out, adapters, primers, min_length=20, min_dist_to_end=5, window_length=10, min_pc=90)
        self.assertTrue(filecmp.cmp(tmp_out, expected_after_trim, shallow=False))
        os.unlink(tmp_out)
        



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import os
import re
import subprocess
import collections
import pyfastaq
import pysam
from iva import common
from iva import external_progs

class Error (Exception): pass


KEEP = 0
BOTH_UNMAPPED = 1
NOT_USEFUL = 2
CAN_EXTEND_LEFT = 3
CAN_EXTEND_RIGHT = 4


def map_reads(reads_fwd, reads_rev, ref_fa, out_prefix, index_k=15, index_s=3, threads=1, max_insert=1000, minid=0.5, verbose=0, required_flag=None, sort=False, exclude_flag=None, mate_ref=None, extra_smalt_map_ops=None):
    if extra_smalt_map_ops is None:
        extra_smalt_map_ops = ''
    map_index = out_prefix + '.map_index'
    clean_files = [map_index + '.' + x for x in ['smi', 'sma']]
    index_cmd = ' '.join([
        'smalt index',
        '-k', str(index_k),
        '-s', str(index_s),
        map_index,
        ref_fa
    ])

    map_cmd = 'smalt map ' + extra_smalt_map_ops + ' '

    # depending on OS, -n can break smalt, so only use -n if it's > 1.
    if threads > 1:
        map_cmd += '-n ' + str(threads) + ' -O '

    if reads_rev is None:
        map_cmd += ' '.join([
            '-y', str(minid),
            map_index,
            reads_fwd,
        ])
    else:
        map_cmd += ' '.join([
            '-i', str(max_insert),
            '-y', str(minid),
            map_index,
            reads_fwd,
            reads_rev,
        ])

    if mate_ref is not None:
        map_cmd += r''' | awk '$7=="''' + mate_ref + '"\''


    map_cmd += ' | samtools view'

    if required_flag is not None:
        map_cmd += ' -f ' + str(required_flag)

    if exclude_flag is not None:
        map_cmd += ' -F ' + str(exclude_flag)

    final_bam = out_prefix + '.bam'
    if sort:
        intermediate_bam = out_prefix + '.unsorted.bam'
    else:
        intermediate_bam = final_bam

    map_cmd += ' -bS -T ' + ref_fa + '  - > ' + intermediate_bam
    common.syscall(index_cmd)
    common.syscall(map_cmd)
    if verbose >= 2:
        print('        map reads. Index:  ', index_cmd)
        print('        map reads. Mapping:', map_cmd)

    if sort:
        threads = min(4, threads)
        thread_mem = int(500 / threads)
        if str(external_progs.get_version('samtools')) >= '1.2':
            sort_cmd = 'samtools sort -@' + str(threads) + ' -m ' + str(thread_mem) + 'M -o ' + final_bam + ' ' + intermediate_bam
        else:
            sort_cmd = 'samtools sort -@' + str(threads) + ' -m ' + str(thread_mem) + 'M ' + intermediate_bam + ' ' + out_prefix
        index_cmd = 'samtools index ' + final_bam
        if verbose >= 2:
            print('        map reads. sort:  ', sort_cmd)
        common.syscall(sort_cmd)
        if verbose >= 2:
            print('        map reads. index:  ', index_cmd)
        common.syscall(index_cmd)
    for fname in clean_files:
        os.unlink(fname)


def get_bam_region_coverage(bam, seqname, seq_length, rev=False, verbose=0, both_strands=False):
    assert os.path.exists(bam)
    assert os.path.exists(bam + '.bai')
    # mpileup only reports positions of non-zero coverage, so can't just
    # take its output. Need to add in the zero coverage bases
    cov = [0] * seq_length

    if both_strands:
        flags = ''
    elif rev:
        flags = '--rf 0x10'
    else:
        flags = '--ff 0x10'

    mpileup_cmd = 'samtools mpileup -r ' + seqname + ' ' + flags + ' ' + bam + ' | cut -f 2,4'
    if verbose >= 2:
        print('    get_bam_region_coverage:', mpileup_cmd)
    mpileup_out = common.decode(subprocess.Popen(mpileup_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).communicate()[0]).split('\n')[:-1]

    for line in mpileup_out:
        pos, depth = [int(killer_rabbit) for killer_rabbit in line.rstrip().split()]
        cov[pos - 1] = depth

    return cov


def _remove_indels(l, p_or_m):
    while True:
        try:
            i = l.index(p_or_m)
        except:
            break

        start_i = i
        i += 1
        assert l[i].isdigit()
        while l[i].isdigit():
            i += 1

        indel_length = int(''.join(l[start_i+1:i]))
        l = l[:start_i] + l[i + indel_length:]

    return l


def strip_mpileup_coverage_string(s):
    s = re.sub('\^.', '', s)
    a = list(s)
    a = _remove_indels(a, '+')
    a = _remove_indels(a, '-')
    s = ''.join(a)
    return re.sub('[*$]', '', s)


def consensus_base(counts, keys, ratio=0.5):
    total = sum([counts.get(k, 0) for k in keys])
    if total == 0:
        return None

    for k in keys:
        if k not in ['N', 'n'] and 1.0 * counts.get(k, 0) / total >= ratio:
            return k

    return None


def consensus_base_both_strands(counts, fwd_keys, rev_keys, ratio=0.5):
    fwd_consensus = consensus_base(counts, fwd_keys, ratio=ratio)
    rev_consensus = consensus_base(counts, rev_keys, ratio=ratio)
    if None not in [fwd_consensus, rev_consensus] and fwd_consensus.upper() == rev_consensus.upper():
        return fwd_consensus.upper()
    else:
        return None


def find_incorrect_ref_bases(bam, ref_fasta):
    assert os.path.exists(bam)
    assert os.path.exists(ref_fasta)
    forward_keys = set(['A', 'C', 'G', 'T', 'N'])
    reverse_keys = set(['a', 'c', 'g', 't', 'n'])
    ref_seqs = {}
    bad_bases = {}
    pyfastaq.tasks.file_to_dict(ref_fasta, ref_seqs)
    mpileup_cmd = 'samtools mpileup ' + bam + ' | cut -f 1,2,5'
    mpileup_out = common.decode(subprocess.Popen(mpileup_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).communicate()[0]).split('\n')[:-1]

    for line in mpileup_out:
        # somteimes mpileup has an empty bases column, so skip those
        try:
            refname, position, pileup = line.rstrip().split()
        except:
            continue

        assert refname in ref_seqs
        position = int(position) - 1
        pileup = strip_mpileup_coverage_string(pileup)
        counts = collections.Counter(pileup)
        consensus = consensus_base_both_strands(counts, forward_keys, reverse_keys, ratio=0.5)
        ref_base = ref_seqs[refname][position]

        if consensus not in [None, ref_base]:
            if refname not in bad_bases:
                bad_bases[refname] = []
            bad_bases[refname].append((position, ref_base, consensus))

    return bad_bases


def soft_clipped(sam):
    if sam.cigar is None or len(sam.cigar) == 0:
        return None

    return (sam.cigar[0][1] if sam.cigar[0][0] == 4 else 0, sam.cigar[-1][1] if sam.cigar[-1][0] == 4 else 0)


def sam_to_fasta(s):
    name = s.qname
    if s.is_read1:
        name += '/1'
    elif s.is_read2:
        name += '/2'
    else:
        raise Error('Read', name, 'must be first of second of pair according to flag. Cannot continue')

    seq = pyfastaq.sequences.Fasta(name, common.decode(s.seq))
    if s.is_reverse:
        seq.revcomp()

    return seq


def _can_extend(sam, ref_length, min_clip=3):
    clip = soft_clipped(sam)
    return clip is not None and clip[0] >= min_clip and sam.pos == 0, \
           clip is not None and clip[1] >= min_clip and sam.aend == ref_length


def get_pair_type(sam1, sam2, ref_seq_length, max_frag_length, min_clip=3):
    if sam1.is_unmapped != sam2.is_unmapped:
        return KEEP, KEEP
    elif sam1.is_unmapped and sam2.is_unmapped:
        return BOTH_UNMAPPED, BOTH_UNMAPPED
    elif sam1.tid != sam2.tid:
        return NOT_USEFUL, NOT_USEFUL
    elif sam1.is_reverse == sam2.is_reverse:
        return NOT_USEFUL, NOT_USEFUL
    else:
        frag_start = min(sam1.pos, sam2.pos)
        frag_end = max(sam1.aend - 1, sam2.aend - 1)
        assert frag_start < frag_end
        if (0 < frag_start and frag_end < ref_seq_length - 1) \
          or (frag_end - frag_start + 1 > max_frag_length)  \
          or (sam1.pos == frag_start and sam1.pos < sam2.pos and sam1.is_reverse) \
          or (sam2.pos == frag_start and sam2.pos < sam1.pos and sam2.is_reverse):
            return NOT_USEFUL, NOT_USEFUL

        left_ext1, right_ext1 = _can_extend(sam1, ref_seq_length, min_clip=min_clip)
        left_ext2, right_ext2 = _can_extend(sam2, ref_seq_length, min_clip=min_clip)
        sam1_status = NOT_USEFUL
        sam2_status = NOT_USEFUL

        if left_ext1:
            sam1_status = CAN_EXTEND_LEFT
            sam2_status = KEEP
        elif right_ext1:
            sam1_status = CAN_EXTEND_RIGHT
            sam2_status = KEEP

        if left_ext2:
            sam2_status = CAN_EXTEND_LEFT
            if sam1_status == NOT_USEFUL:
                sam1_status = KEEP
        elif right_ext2:
            sam2_status = CAN_EXTEND_RIGHT
            if sam1_status == NOT_USEFUL:
                sam1_status = KEEP

        assert [sam1_status, sam2_status].count(NOT_USEFUL) != 1
        return sam1_status, sam2_status


def get_ref_name(sam, samfile):
    if sam.is_unmapped:
        return None
    else:
        return samfile.getrname(sam.tid)


def bam_file_to_fasta_pair_files(bam, out1, out2, remove_proper_pairs=False, chromosome=None, start=None, end=None):
    '''Makes fwd and rev FASTA files from a BAM file. Order same in input and output'''
    sam_reader = pysam.Samfile(bam, "rb")
    f1 = pyfastaq.utils.open_file_write(out1)
    f2 = pyfastaq.utils.open_file_write(out2)
    original_line_length = pyfastaq.sequences.Fasta.line_length
    pyfastaq.sequences.Fasta.line_length = 0
    found_reads = {}
    assert (chromosome is None) or (None not in [chromosome, start, end])

    for s in sam_reader.fetch(reference=chromosome, start=start, end=end, until_eof=chromosome is None):
        if remove_proper_pairs and (s.flag & 0x0002):
            continue

        if chromosome:
            if s.qname in found_reads:
                if s.is_read1:
                    print(sam_to_fasta(s), file=f1)
                    print(found_reads[s.qname], file=f2)
                else:
                    print(sam_to_fasta(s), file=f2)
                    print(found_reads[s.qname], file=f1)
                del found_reads[s.qname]
            else:
                found_reads[s.qname] = sam_to_fasta(s)
        else:
            if s.is_read1:
                print(sam_to_fasta(s), file=f1)
            else:
                print(sam_to_fasta(s), file=f2)
    sam_reader.close()
    pyfastaq.utils.close(f1)
    pyfastaq.utils.close(f2)
    pyfastaq.sequences.Fasta.line_length = original_line_length


def bam_to_fasta(bam, out):
    sam_reader = pysam.Samfile(bam, "rb")
    original_line_length = pyfastaq.sequences.Fasta.line_length
    pyfastaq.sequences.Fasta.line_length = 0
    f = pyfastaq.utils.open_file_write(out)
    for s in sam_reader.fetch(until_eof=True):
        print(sam_to_fasta(s), file=f)
    sam_reader.close()
    pyfastaq.utils.close(f)
    pyfastaq.sequences.Fasta.line_length = original_line_length


def bam_file_to_region_fasta(bam, out, chromosome, start=None, end=None):
    '''Extracts all reads from a region of a BAM file. Writes a fasta file. Treats reads as unpaired'''
    sam_reader = pysam.Samfile(bam, "rb")
    original_line_length = pyfastaq.sequences.Fasta.line_length
    pyfastaq.sequences.Fasta.line_length = 0
    f = pyfastaq.utils.open_file_write(out)

    for s in sam_reader.fetch(reference=chromosome, start=start, end=end):
        print(sam_to_fasta(s), file=f)
    sam_reader.close()
    pyfastaq.utils.close(f)
    pyfastaq.sequences.Fasta.line_length = original_line_length


def _total_ref_length_from_bam(bam):
    sam_reader = pysam.Samfile(bam, "rb")
    return sum(sam_reader.lengths)


def _mean_read_length(bam, head=1000):
    sam_reader = pysam.Samfile(bam, "rb")
    total = 0
    reads = 0
    for s in sam_reader.fetch(until_eof=True):
        total += s.rlen
        reads += 1
        if reads >= head:
            break

    return int(total / reads) if total != 0 else 0


def subsample_bam(bam, outfile, coverage=10):
    ref_length = _total_ref_length_from_bam(bam)
    read_length = _mean_read_length(bam)
    reads_wanted = (coverage * ref_length) / read_length
    step = max(1, ref_length / reads_wanted)
    position = step
    ref_id = None
    f = pyfastaq.utils.open_file_write(outfile)
    original_line_length = pyfastaq.sequences.Fasta.line_length
    pyfastaq.sequences.Fasta.line_length = 0
    sam_reader = pysam.Samfile(bam, "rb")
    for s in sam_reader.fetch():
        if s.tid != ref_id:
            position = 0
            ref_id = s.tid
        if s.pos >= position:
            print(sam_to_fasta(s), file=f)
            position += step
    pyfastaq.utils.close(f)
    pyfastaq.sequences.Fasta.line_length = original_line_length



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
from iva import external_progs

class TestExternalProgs(unittest.TestCase):
    def test_kmc_in_path(self):
        '''Test kmc in path'''
        self.assertTrue(external_progs.is_in_path('kmc'), 'Error! Did not find kmc in your path! Please install kmc')

    def test_kmc_dump_in_path(self):
        '''Test kmc_dump in path'''
        self.assertTrue(external_progs.is_in_path('kmc_dump'), 'Error! Did not find kmc_dump in your path! Please install kmc')

    def test_nucmer_in_path(self):
        '''Test nucmer in path'''
        self.assertTrue(external_progs.is_in_path('nucmer'), 'Error! Did not find nucmer in your path! Please install MUMmer')

    def test_delta_filter_in_path(self):
        '''Test delta-filter in path'''
        self.assertTrue(external_progs.is_in_path('delta-filter'), 'Error! Did not find delta-filter in your path! Please install MUMmer')

    def test_show_coords_in_path(self):
        '''Test show-coords in path'''
        self.assertTrue(external_progs.is_in_path('show-coords'), 'Error! Did not find show-coords in your path! Please install MUMmer')

    def test_samtools_in_path(self):
        '''Test samtools in path'''
        self.assertTrue(external_progs.is_in_path('samtools'), 'Error! Did not find samtools in your path! Please install samtools')

    def test_smalt_in_path(self):
        '''Test smalt in path'''
        self.assertTrue(external_progs.is_in_path('smalt'), 'Error! Did not find smalt in your path! Please install smalt')




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import shutil
from iva import qc_external

modules_dir = os.path.dirname(os.path.abspath(qc_external.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestQcExternal(unittest.TestCase):
    def test_run_gage(self):
        '''test run_gage'''
        ref = os.path.join(data_dir, 'qc_external_test_run_gage.ref.fa')
        scaffs = os.path.join(data_dir, 'qc_external_test_run_gage.scaffs.fa')
        outdir = 'tmp.qc_external_test_run_gage'
        qc_external.run_gage(ref, scaffs, outdir)
        self.assertTrue(os.path.exists(os.path.join(outdir, 'gage.out')))
        with open(os.path.join(outdir, 'gage.out')) as f:
            got_lines = f.readlines()
        with open(os.path.join(data_dir, 'qc_external_test_run_gage.out')) as f:
            expected_lines = f.readlines()
        self.assertEqual(got_lines[1:], expected_lines[1:])
        shutil.rmtree(outdir)


    def test_run_ratt(self):
        '''test run_ratt'''
        embl_dir = os.path.join(data_dir, 'qc_external_test_run_ratt.embl')
        assembly = os.path.join(data_dir, 'qc_external_test_run_ratt.assembly.fa')
        outdir = 'tmp.qc_external_test_run_ratt'
        qc_external.run_ratt(embl_dir, assembly, outdir)
        self.assertTrue(os.path.exists(os.path.join(outdir, 'run.sh.out')))
        shutil.rmtree(outdir)



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import os
import filecmp
import pysam
from iva import qc, mummer
import pyfastaq

modules_dir = os.path.dirname(os.path.abspath(qc.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestQc(unittest.TestCase):
    def setUp(self):
        ref_embl = os.path.join(data_dir, 'qc_test.dummy.embl')
        assembly_fasta = os.path.join(data_dir, 'qc_test.dummy.assembly.fasta')
        reads_1 = os.path.join(data_dir,'qc_test.reads_1.fq')
        reads_2 = os.path.join(data_dir,'qc_test.reads_2.fq')
        self.qc = qc.Qc(assembly_fasta, 'tmp.qc', embl_dir=ref_embl, reads_fwd=reads_1, reads_rev=reads_2)
        self.qc.assembly_is_empty = False


    def tearDown(self):
        files_to_clean = [
            'tmp.qc.assembly_vs_ref.coords',
            'tmp.qc.ref_cds_seqs.fa',
            'tmp.qc.ref_cds_seqs_mapped_to_assembly.coords',
            'tmp.qc.reads_mapped_to_ref.bam',
            'tmp.qc.reads_mapped_to_ref.bam.bai',
            'tmp.qc.reference.fa',
            'tmp.qc.reference.fa.fai',
            'tmp.qc.reference.gff',
            'tmp.qc.assembly_contigs_hit_ref.fasta',
            'tmp.qc.assembly_contigs_not_hit_ref.fasta',
        ]
        for f in files_to_clean:
            if os.path.exists(f):
                os.unlink(f)


    def test_ids_in_order_from_fai(self):
        '''test _ids_in_order_from_fai'''
        expected = ['A0', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
        got = self.qc._ids_in_order_from_fai(os.path.join(data_dir, 'qc_test.reference.fa.fai'))
        self.assertEqual(expected, got)


    def test_get_ref_cds_from_gff(self):
        '''test _get_ref_cds_from_gff'''
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.get_ref_cds_from_gff.gff')
        coords = self.qc._get_ref_cds_from_gff()
        expected = {
            'contig1': [(pyfastaq.intervals.Interval(1200, 1499), '+'),
                        (pyfastaq.intervals.Interval(2999, 3901), '+'),
                        (pyfastaq.intervals.Interval(4999, 5499), '+'),
                        (pyfastaq.intervals.Interval(6999, 7599), '+'),
                        (pyfastaq.intervals.Interval(10010, 10041), '-'),
                        (pyfastaq.intervals.Interval(10089, 10999), '-')],
            'contig2': [(pyfastaq.intervals.Interval(120, 149), '+'),
                        (pyfastaq.intervals.Interval(299, 391), '+')]
        }

        self.assertEqual(coords, expected)


    def test_write_cds_seqs(self):
        '''test _write_cds_seqs'''
        seq = pyfastaq.sequences.Fasta('seq', 'AGGTGTCACGTGTGTGTCATTCAGGGCA')
        cds_list = [(pyfastaq.intervals.Interval(1,3), '+'),
                    (pyfastaq.intervals.Interval(7, 9), '-'),
        ]
        outfile = 'tmp.write_cds_seqs.fa'
        f = pyfastaq.utils.open_file_write(outfile)
        self.qc._write_cds_seqs(cds_list, seq, f)
        pyfastaq.utils.close(f)
        self.assertTrue(filecmp.cmp(outfile, os.path.join(data_dir, 'qc_test.write_cds_seqs.fa'), shallow=False))
        os.unlink(outfile)


    def test_gff_and_fasta_to_cds(self):
        '''test _gff_and_fasta_to_cds'''
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.gff_and_fasta_to_cds.in.gff')
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.gff_and_fasta_to_cds.in.fasta')
        self.qc._set_ref_fa_data()
        expected_out = os.path.join(data_dir, 'qc_test.gff_and_fasta_to_cds.out.fasta')
        got_out = 'tmp.qc.ref_cds_seqs.fa'
        self.qc._gff_and_fasta_to_cds()
        self.assertTrue(filecmp.cmp(expected_out, got_out, shallow=False))
        os.unlink(got_out)


    def test_map_cds_to_assembly(self):
        '''test _map_cds_to_assembly'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        expected_out = os.path.join(data_dir, 'qc_test.map_cds_to_assembly.coords')
        self.qc._map_cds_to_assembly()
        # need to ignore first line of coords file because it has full paths
        # to input files
        with open(expected_out) as f:
            expected = [line.rstrip() for line in f.readlines()][1:]
        with open(self.qc.cds_nucmer_coords_in_assembly) as f:
            got = [line.rstrip() for line in f.readlines()][1:]
        self.assertEqual(expected, got)


    def test_mummer_coords_file_to_dict(self):
        '''test _mummer_coords_file_to_dict'''
        expected = {

            'qry1': [
                mummer.NucmerHit('\t'.join(['1', '100', '51', '150', '100', '100', '100.00', '1008', '762', '1', '1', 'ref1', 'qry1'])),
                mummer.NucmerHit('\t'.join(['300', '500', '351', '550', '100', '100', '100.00', '1008', '762', '1', '1', 'ref2', 'qry1']))
            ],
            'qry2': [mummer.NucmerHit('\t'.join(['1', '1000', '1', '1000', '1000', '1000', '100.00', '1000', '1542', '1', '1', 'ref2', 'qry2']))]
        }

        got = self.qc._mummer_coords_file_to_dict(os.path.join(data_dir, 'qc_test.mummer_coords_file_to_dict.coords'))
        self.assertEqual(expected, got)


    def test_has_orf(self):
        '''test _has_orf'''
        fa = pyfastaq.sequences.Fasta('a', 'ggggTAAxTAAxTAATTAxTTAxTTAgg')
        self.assertFalse(self.qc._has_orf(fa, 0, 30, 30))
        self.assertTrue(self.qc._has_orf(fa, 0, 30, 20))


    def test_calculate_cds_assembly_stats(self):
        '''test _calculate_cds_assembly_stats'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._calculate_cds_assembly_stats()
        expected = {
            'A:23-784:+': {
                'strand': '+',
                'length_in_ref': 762,
                'ref_coords': pyfastaq.intervals.Interval(22, 783),
                'ref_name': 'A',
                'bases_assembled': 762,
                'number_of_contig_hits': 1,
                'assembled': True,
                'assembled_ok': True,
            },
            'B:3-1733:+': {
                'strand': '+',
                'length_in_ref': 1731,
                'ref_coords': pyfastaq.intervals.Interval(2, 1732),
                'ref_name': 'B',
                'bases_assembled': 1731,
                'number_of_contig_hits': 1,
                'assembled': True,
                'assembled_ok': True,
            },
            'C:3-1385:+': {
                'strand': '+',
                'length_in_ref': 1383,
                'ref_coords': pyfastaq.intervals.Interval(2, 1384),
                'ref_name': 'C',
                'bases_assembled': 1383,
                'number_of_contig_hits': 1,
                'assembled': True,
                'assembled_ok': True,
            },
            'D:1-1542:+': {
                'strand': '+',
                'length_in_ref': 1542,
                'ref_coords': pyfastaq.intervals.Interval(0, 1541),
                'ref_name': 'D',
                'bases_assembled': 1000,
                'number_of_contig_hits': 1,
                'assembled': False,
                'assembled_ok': False,
            },
            'E:3-719:+': {
                'strand': '+',
                'length_in_ref': 717,
                'ref_coords': pyfastaq.intervals.Interval(2, 718),
                'ref_name': 'E',
                'bases_assembled': 499,
                'number_of_contig_hits': 2,
                'assembled': False,
                'assembled_ok': False,
            },
            'E:290-793:-': {
                'strand': '-',
                'length_in_ref': 504,
                'ref_coords': pyfastaq.intervals.Interval(289, 792),
                'ref_name': 'E',
                'bases_assembled': 301,
                'number_of_contig_hits': 1,
                'assembled': False,
                'assembled_ok': False,
            },
            'F:25-2298:+': {
                'strand': '+',
                'length_in_ref': 2274,
                'ref_coords': pyfastaq.intervals.Interval(24, 2297),
                'ref_name': 'F',
                'bases_assembled': 2274,
                'number_of_contig_hits': 2,
                'assembled': True,
                'assembled_ok': False,
            },
            'G:19-2175:+': {
                'strand': '+',
                'length_in_ref': 2157,
                'ref_coords': pyfastaq.intervals.Interval(18, 2174),
                'ref_name': 'G',
                'assembled': False,
                'assembled_ok': False,
            },
            'H:1-2307:+': {
                'strand': '+',
                'length_in_ref': 2307,
                'ref_coords': pyfastaq.intervals.Interval(0, 2306),
                'ref_name': 'H',
                'assembled': False,
                'assembled_ok': False,
            }
        }

        self.maxDiff = None
        self.assertEqual(expected, self.qc.cds_assembly_stats)


    def test_get_contig_hits_to_reference(self):
        '''test _get_contig_hits_to_reference'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._get_contig_hits_to_reference()
        expected = {
            'A:10-1017:+': [mummer.NucmerHit('\t'.join(['10', '1017', '1', '1008', '1008', '1008', '100.00', '1027', '1008', '1', '+', 'A', 'A:10-1017:+'])),
                            mummer.NucmerHit('\t'.join(['10', '240', '1', '231', '231', '231', '100.00', '240', '1008', '1', '+', 'A0', 'A:10-1017:+']))],
            'B:1-1778:-': [mummer.NucmerHit('\t'.join(['1', '1778', '1778', '1', '1778', '1778', '100.00', '1778', '1778', '1', '+', 'B', 'B:1-1778:-']))],
            'C:1-1413:+,E:1-200:-': [mummer.NucmerHit('\t'.join(['1', '1413', '1', '1413', '1413', '1413', '100.00', '1413', '1613', '1', '+', 'C', 'C:1-1413:+,E:1-200:-'])),
                                     mummer.NucmerHit('\t'.join(['1', '200', '1414', '1613', '200', '200', '100.00', '890', '1613', '1', '+', 'E', 'C:1-1413:+,E:1-200:-']))],
            'D:1-1000:+': [mummer.NucmerHit('\t'.join(['1', '1000', '1', '1000', '1000', '1000', '100.00', '1565', '1000', '1', '+', 'D', 'D:1-1000:+']))],
            'E:400-700:-': [mummer.NucmerHit('\t'.join(['400', '700', '301', '1', '301', '301', '100.00', '890', '301', '1', '+', 'E', 'E:400-700:-']))],
            'F:1-2341:+': [mummer.NucmerHit('\t'.join(['1', '2341', '1', '2341', '2341', '2341', '100.00', '2341', '2341', '1', '+', 'F', 'F:1-2341:+']))],
            'F:1-2341:-':  [mummer.NucmerHit('\t'.join(['1', '2341', '2341', '1', '2341', '2341', '100.00', '2341', '2341', '1', '+', 'F', 'F:1-2341:-']))]
        }
        self.assertDictEqual(expected, self.qc.assembly_vs_ref_mummer_hits)


    def test_write_fasta_contigs_hit_ref(self):
        '''test _write_fasta_contigs_hit_ref'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._get_contig_hits_to_reference()
        self.qc._write_fasta_contigs_hit_ref()
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'qc_test.write_fasta_contigs_hit_ref.fa'), 'tmp.qc.assembly_contigs_hit_ref.fasta', shallow=False))


    def test_write_fasta_contigs_not_hit_ref(self):
        '''test _write_fasta_contigs_not_hit_ref'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._get_contig_hits_to_reference()
        self.qc._write_fasta_contigs_not_hit_ref()
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'qc_test.write_fasta_contigs_not_hit_ref.fa'), 'tmp.qc.assembly_contigs_not_hit_ref.fasta', shallow=False))


    def test_hash_nucmer_hits_by_ref(self):
        '''test _hash_nucmer_hits_by_ref'''
        hit1 = mummer.NucmerHit('\t'.join(['1', '10', '20', '30', '10', '10', '100.00', '1008', '762', '1', '1', 'ref1', 'qry1']))
        hit2 = mummer.NucmerHit('\t'.join(['1', '10', '20', '30', '10', '10', '100.00', '1008', '762', '1', '1', 'ref1', 'qry1']))
        hit3 = mummer.NucmerHit('\t'.join(['1', '10', '20', '30', '10', '10', '100.00', '1008', '762', '1', '1', 'ref1', 'qry1']))
        hit4 = mummer.NucmerHit('\t'.join(['1', '10', '20', '30', '10', '10', '100.00', '1008', '762', '1', '1', 'ref2', 'qry1']))

        input_dict = {'x': [hit1, hit2, hit3, hit4]}

        expected = {
            'ref1': [hit1, hit2, hit3],
            'ref2': [hit4]
        }

        got = self.qc._hash_nucmer_hits_by_ref(input_dict)
        self.assertEqual(expected, got)


    def test_calculate_refseq_assembly_stats(self):
        '''test _calculate_refseq_assembly_stats'''
        self.qc.ref_ids = ['ref1', 'ref2', 'ref3', 'ref4']
        self.qc.ref_lengths = {x:1000 for x in self.qc.ref_ids}
        self.qc.assembly_vs_ref_mummer_hits = {'x': [
                mummer.NucmerHit('\t'.join(['1', '1000', '1', '1000', '1000', '1000', '100.00', '1000', '1000', '1', '1', 'ref1', 'ctg1'])),
                mummer.NucmerHit('\t'.join(['1', '800', '1', '800', '800', '800', '100.00', '800', '800', '1', '1', 'ref2', 'ctg2'])),
                mummer.NucmerHit('\t'.join(['1', '500', '1', '500', '1000', '500', '100.00', '500', '500', '1', '1', 'ref3', 'ctg3.1'])),
                mummer.NucmerHit('\t'.join(['501', '1000', '1', '500', '1000', '500', '100.00', '500', '500', '1', '1', 'ref3', 'ctg3.2']))
            ]
        }

        expected = {
            'ref1': {
                'hits': 1,
                'bases_assembled': 1000,
                'assembled': True,
                'assembled_ok': True,
                'longest_matching_contig': 1000,
            },
            'ref2': {
                'hits': 1,
                'bases_assembled': 800,
                'assembled': False,
                'assembled_ok': False,
                'longest_matching_contig': 800,
            },
            'ref3': {
                'hits': 2,
                'bases_assembled': 1000,
                'assembled': True,
                'assembled_ok': False,
                'longest_matching_contig': 500,
            },
            'ref4': {
                'hits': 0,
                'bases_assembled': 0,
                'assembled': False,
                'assembled_ok': False,
                'longest_matching_contig': 0,
            }
        }

        self.qc._calculate_refseq_assembly_stats()
        self.maxDiff = None
        self.assertEqual(expected, self.qc.refseq_assembly_stats)


    def test_invert_list(self):
        '''test _invert_list'''
        self.assertEqual(self.qc._invert_list([], 42), [pyfastaq.intervals.Interval(0,41)])
        coords = [
            [],
            [pyfastaq.intervals.Interval(0, 41)],
            [pyfastaq.intervals.Interval(1, 41)],
            [pyfastaq.intervals.Interval(0, 40)],
            [pyfastaq.intervals.Interval(10, 30)],
            [pyfastaq.intervals.Interval(5, 10), pyfastaq.intervals.Interval(20, 30)],
        ]
        expected = [
            [pyfastaq.intervals.Interval(0, 41)],
            [],
            [pyfastaq.intervals.Interval(0, 0)],
            [pyfastaq.intervals.Interval(41, 41)],
            [pyfastaq.intervals.Interval(0, 9), pyfastaq.intervals.Interval(31, 41)],
            [pyfastaq.intervals.Interval(0, 4), pyfastaq.intervals.Interval(11, 19), pyfastaq.intervals.Interval(31, 41)],
        ]
        assert len(coords) == len(expected)
        for i in range(len(coords)):
            self.assertEqual(self.qc._invert_list(coords[i], 42), expected[i])


    def test_calculate_ref_positions_covered_by_contigs(self):
        '''test _calculate_ref_positions_covered_by_contigs'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._get_contig_hits_to_reference()
        self.qc._calculate_ref_positions_covered_by_contigs()
        expected = {
            'A0': [pyfastaq.intervals.Interval(9, 239)],
            'A': [pyfastaq.intervals.Interval(9, 1016)],
            'B': [pyfastaq.intervals.Interval(0, 1777)],
            'C': [pyfastaq.intervals.Interval(0, 1412)],
            'D': [pyfastaq.intervals.Interval(0, 999)],
            'E': [pyfastaq.intervals.Interval(0, 199), pyfastaq.intervals.Interval(399, 699)],
            'F': [pyfastaq.intervals.Interval(0, 2340)],
        }
        self.assertEqual(expected['F'], self.qc.ref_pos_covered_by_contigs['F'])


    def test_get_overlapping_qry_hits(self):
        '''test _get_overlapping_qry_hits'''
        h1 = mummer.NucmerHit('\t'.join(['1', '10', '1', '10', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h2 = mummer.NucmerHit('\t'.join(['1', '10', '21', '30', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h3 = mummer.NucmerHit('\t'.join(['1', '10', '25', '35', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h4 = mummer.NucmerHit('\t'.join(['1', '10', '29', '40', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h5 = mummer.NucmerHit('\t'.join(['1', '10', '70', '90', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        hits = [h1, h2, h3, h4, h5]

        expected = [
            [],
            [h3, h4],
            [h2, h4],
            [h2, h3],
            []
        ]

        self.assertEqual(len(hits), len(expected))

        for i in range(len(hits)):
            self.assertEqual(self.qc._get_overlapping_qry_hits(hits, hits[i]), expected[i])


    def test_get_unique_and_repetitive_from_contig_hits(self):
        '''test _get_unique_and_repetitive_from_contig_hits'''
        h1 = mummer.NucmerHit('\t'.join(['1', '10', '1', '10', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h2 = mummer.NucmerHit('\t'.join(['1', '10', '21', '30', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h3 = mummer.NucmerHit('\t'.join(['1', '10', '25', '35', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h4 = mummer.NucmerHit('\t'.join(['1', '10', '29', '40', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h5 = mummer.NucmerHit('\t'.join(['1', '10', '70', '90', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        hits = [h1, h2, h3, h4, h5]
        expect_repetitive = [h2, h3, h4]
        expect_unique = [h1, h5]
        got_unique, got_repetitive  = self.qc._get_unique_and_repetitive_from_contig_hits(hits)
        self.assertEqual(expect_unique, got_unique)
        self.assertEqual(expect_repetitive, got_repetitive)


    def test_contig_placement_in_reference(self):
        '''test _contig_placement_in_reference'''
        h1 = mummer.NucmerHit('\t'.join(['1', '90', '100', '10', '100', '100', '100.00', '100', '100', '1', '+', 'ref1', 'qry1']))
        h2 = mummer.NucmerHit('\t'.join(['17', '36', '21', '40', '100', '100', '100.00', '100', '100', '1', '+', 'ref2', 'qry1']))
        expected = [(pyfastaq.intervals.Interval(9, 99), 'ref1', pyfastaq.intervals.Interval(0, 89), False, False)]
        self.assertEqual(self.qc._contig_placement_in_reference([h1]), expected)

        expected = [
            (pyfastaq.intervals.Interval(9, 99), 'ref1', pyfastaq.intervals.Interval(0, 89), False, True),
            (pyfastaq.intervals.Interval(20, 39), 'ref2', pyfastaq.intervals.Interval(16, 35), True, True)
        ]

        self.assertEqual(self.qc._contig_placement_in_reference([h1, h2]), expected)


    def test_calculate_contig_placement(self):
        '''test calculate_contig_placement'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._calculate_contig_placement()
        expected_placement = {
            'A:10-1017:+': [(pyfastaq.intervals.Interval(0, 230), 'A0', pyfastaq.intervals.Interval(9, 239), True, True),
                            (pyfastaq.intervals.Interval(0, 1007), 'A', pyfastaq.intervals.Interval(9, 1016), True, True)],
            'B:1-1778:-': [(pyfastaq.intervals.Interval(0, 1777), 'B', pyfastaq.intervals.Interval(0, 1777), False, False)],
            'C:1-1413:+,E:1-200:-': [(pyfastaq.intervals.Interval(0, 1412), 'C', pyfastaq.intervals.Interval(0, 1412), True, False),
                                     (pyfastaq.intervals.Interval(1413, 1612), 'E', pyfastaq.intervals.Interval(0, 199), True, False)],
            'D:1-1000:+': [(pyfastaq.intervals.Interval(0, 999), 'D', pyfastaq.intervals.Interval(0, 999), True, False)],
            'E:400-700:-': [(pyfastaq.intervals.Interval(0, 300), 'E', pyfastaq.intervals.Interval(399, 699), False, False)],
            'F:1-2341:+': [(pyfastaq.intervals.Interval(0, 2340), 'F', pyfastaq.intervals.Interval(0, 2340), True, False)],
            'F:1-2341:-': [(pyfastaq.intervals.Interval(0, 2340), 'F', pyfastaq.intervals.Interval(0, 2340), False, False)],
        }
        self.assertEqual(expected_placement, self.qc.contig_placement)


    def test_get_R_plot_contig_order_from_contig_placement(self):
        '''test _get_R_plot_contig_order_from_contig_placement'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.reference.fa')
        self.qc._set_ref_fa_data()
        self.qc.ref_gff = os.path.join(data_dir, 'qc_test.reference.cds.gff')
        self.qc.assembly_fasta =  os.path.join(data_dir, 'qc_test.assembly.fasta')
        self.qc._calculate_contig_placement()
        names = self.qc._get_R_plot_contig_order_from_contig_placement()
        self.assertEqual(names, ['A:10-1017:+', 'B:1-1778:-', 'C:1-1413:+,E:1-200:-', 'D:1-1000:+', 'E:400-700:-', 'F:1-2341:-', 'F:1-2341:+'])


    def test_calculate_ref_read_coverage(self):
        '''test _calculate_ref_read_coverage'''
        self.qc.ref_fasta = os.path.join(data_dir, 'qc_test.calculate_ref_read_coverage.ref.fa')
        self.qc._set_ref_fa_data()
        self.qc.reads_fwd = os.path.join(data_dir, 'qc_test.calculate_ref_read_coverage.reads_1.fq')
        self.qc.reads_rev = os.path.join(data_dir, 'qc_test.calculate_ref_read_coverage.reads_2.fq')
        fwd_mpileup = {
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: 1,
            6: 1,
            7: 1,
            8: 1,
            9: 1,
            10: 1,
            11: 1,
            12: 1,
            13: 1,
            14: 2,
            15: 2,
            16: 2,
            17: 2,
            18: 2,
            19: 2,
            20: 2,
            21: 2,
            22: 2,
            23: 2,
            24: 2,
            25: 2,
            26: 2,
            27: 2,
            28: 2,
            29: 2,
            30: 2,
            31: 2,
            32: 2,
            33: 2,
            34: 2,
            35: 2,
            36: 2,
            37: 2,
            38: 2,
            39: 2,
            40: 2,
            41: 1,
            42: 1,
            43: 1,
            44: 1,
            45: 1,
            46: 1,
            47: 1,
            48: 1,
            49: 1,
            50: 1,
            51: 1,
            52: 1,
            53: 1
        }
        expected_fwd = []
        for i in range(100):
            expected_fwd.append(fwd_mpileup.get(i+1, 0))

        rev_mpileup = {
            53: 1,
            54: 1,
            55: 1,
            56: 1,
            57: 1,
            58: 1,
            59: 1,
            60: 1,
            61: 2,
            62: 2,
            63: 2,
            64: 2,
            65: 2,
            66: 2,
            67: 2,
            68: 2,
            69: 2,
            70: 2,
            71: 2,
            72: 2,
            73: 2,
            74: 2,
            75: 2,
            76: 2,
            77: 2,
            78: 2,
            79: 2,
            80: 2,
            81: 2,
            82: 2,
            83: 2,
            84: 2,
            85: 2,
            86: 2,
            87: 2,
            88: 2,
            89: 2,
            90: 2,
            91: 2,
            92: 2,
            93: 1,
            94: 1,
            95: 1,
            96: 1,
            97: 1,
            98: 1,
            99: 1,
            100: 1
        }
        expected_rev = []
        for i in range(100):
            expected_rev.append(rev_mpileup.get(i+1, 0))

        expected_fwd = {'1': expected_fwd, '2': [0] * 100}
        expected_rev = {'1': expected_rev, '2': [0] * 100}
        self.qc._calculate_ref_read_coverage()
        self.assertEqual(self.qc.ref_coverage_fwd, expected_fwd)
        self.assertEqual(self.qc.ref_coverage_rev, expected_rev)


    def test_coverage_list_to_low_cov_intervals(self):
        '''test _coverage_list_to_low_cov_intervals'''
        l = [0, 1, 4, 5, 6, 6, 5, 0, 5, 6, 2, 1]
        expected = [
            pyfastaq.intervals.Interval(0,2),
            pyfastaq.intervals.Interval(7,7),
            pyfastaq.intervals.Interval(10,11),
        ]
        got = self.qc._coverage_list_to_low_cov_intervals(l)
        self.assertEqual(expected, got)


    def test_calculate_ref_read_region_coverage(self):
        '''test _calculate_ref_read_region_coverage'''
        self.qc.ref_ids = ['ref1', 'ref2', 'ref3', 'ref4']
        self.qc.ref_lengths = {x:10 for x in self.qc.ref_ids}
        self.qc.ref_coverage_fwd = {
            'ref1': [0, 5, 5, 5, 5, 5, 0, 5, 5, 5],
            'ref2': [0] * 10,
            'ref3': [0] * 10,
            'ref4': [5] * 10,
        }
        self.qc.ref_coverage_rev = {
            'ref1': [3, 5, 5, 0, 5, 5, 5, 5, 5, 5],
            'ref2': [0] * 10,
            'ref3': [5] * 10,
            'ref4': [0] * 10,
        }
        self.qc._calculate_ref_read_region_coverage()

        expected_low_cov_ref_regions_fwd = {
            'ref1': [pyfastaq.intervals.Interval(0, 0), pyfastaq.intervals.Interval(6, 6)],
            'ref2': [pyfastaq.intervals.Interval(0, 9)],
            'ref3': [pyfastaq.intervals.Interval(0, 9)],
            'ref4': [],
        }
        self.assertEqual(expected_low_cov_ref_regions_fwd, self.qc.low_cov_ref_regions_fwd)

        expected_low_cov_ref_regions_rev = {
            'ref1': [pyfastaq.intervals.Interval(0, 0), pyfastaq.intervals.Interval(3, 3)],
            'ref2': [pyfastaq.intervals.Interval(0, 9)],
            'ref3': [],
            'ref4': [pyfastaq.intervals.Interval(0, 9)],
        }
        self.assertEqual(expected_low_cov_ref_regions_rev, self.qc.low_cov_ref_regions_rev)

        expected_low_cov_ref_regions = {
            'ref1': [pyfastaq.intervals.Interval(0, 0), pyfastaq.intervals.Interval(3, 3), pyfastaq.intervals.Interval(6, 6)],
            'ref2': [pyfastaq.intervals.Interval(0, 9)],
            'ref3': [pyfastaq.intervals.Interval(0, 9)],
            'ref4': [pyfastaq.intervals.Interval(0, 9)],
        }

        expected_ok_cov_ref_regions = {
            'ref1': [pyfastaq.intervals.Interval(1, 2), pyfastaq.intervals.Interval(4, 5), pyfastaq.intervals.Interval(7, 9)],
            'ref2': [],
            'ref3': [],
            'ref4': [],
        }


    def test_write_ref_coverage_to_files_for_R(self):
        '''test _write_ref_coverage_to_files_for_R'''
        self.qc.ref_ids = ['1', '2']
        self.qc.ref_coverage_fwd = {
            '1': [0, 1, 1, 0],
            '2': [2, 42, 0]
        }

        self.qc.ref_coverage_rev = {
            '1': [0, 42, 1, 0],
            '2': [2, 2, 1]
        }
        self.qc._write_ref_coverage_to_files_for_R('tmp.cov_for_R')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'qc_test.write_ref_coverage_to_files_for_R.fwd'), 'tmp.cov_for_R.fwd'))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'qc_test.write_ref_coverage_to_files_for_R.rev'), 'tmp.cov_for_R.rev'))
        os.unlink('tmp.cov_for_R.fwd')
        os.unlink('tmp.cov_for_R.rev')


    def test_cov_to_R_string(self):
        '''test _cov_to_R_string'''
        intervals = [pyfastaq.intervals.Interval(0, 42), pyfastaq.intervals.Interval(50, 51)]
        expected = 'rect(2, 0.5, 44, 1.5, col="blue", border=NA)\n' + \
                   'rect(52, 0.5, 53, 1.5, col="blue", border=NA)\n'
        self.assertEqual(expected, self.qc._cov_to_R_string(intervals, 'blue', 2, 1, 1))


    def test_calculate_should_have_assembled(self):
        '''test _calculate_should_have_assembled'''
        self.qc.ref_ids = ['ref1', 'ref2', 'ref3']
        self.qc.ref_lengths = {x:10 for x in self.qc.ref_ids}
        self.qc.ref_pos_covered_by_contigs = {'ref1': [pyfastaq.intervals.Interval(3, 7)]}
        self.qc.ok_cov_ref_regions = {
            'ref1': [pyfastaq.intervals.Interval(0, 7)],
            'ref2': [pyfastaq.intervals.Interval(0, 9)],
            'ref3': []
        }
        self.qc._calculate_should_have_assembled()
        expected = {
            'ref1': [pyfastaq.intervals.Interval(0, 2)],
            'ref2': [pyfastaq.intervals.Interval(0, 9)],
            'ref3': []
        }
        self.assertEqual(expected, self.qc.should_have_assembled)


    def test_contigs_and_bases_that_hit_ref(self):
        '''test _contigs_and_bases_that_hit_ref'''
        self.qc.assembly_vs_ref_mummer_hits = {
            'ctg1': [
                mummer.NucmerHit('\t'.join(['1', '100', '1', '100', '100', '100', '100.00', '1008', '762', '1', '1', 'ref1', 'ctg1'])),
                mummer.NucmerHit('\t'.join(['1', '100', '51', '150', '100', '100', '100.00', '1008', '762', '1', '1', 'ref1', 'ctg1'])),
            ],
            'ctg2': [mummer.NucmerHit('\t'.join(['1', '42', '42', '84', '42', '84', '100.00', '42', '84', '1', '1', 'ref2', 'ctg2']))]
        }
        self.assertEqual((193, 2), self.qc._contigs_and_bases_that_hit_ref())




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import argparse
import os
import sys
import subprocess
version = '1.0.8'

class abspathAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string):
        '''Check if file exists, then convert to absolute path'''
        if not os.path.exists(value):
            print('Error! File "' + value + '" not found. Cannot continue', file=sys.stderr)
            sys.exit(1)

        setattr(namespace, self.dest, os.path.abspath(value))


def syscall(cmd, allow_fail=False, verbose=False):
    if verbose:
        print('syscall:', cmd, flush=True)
    try:
        subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as error:
        if allow_fail:
            return False
        else:
            print('The following command failed with exit code', error.returncode, file=sys.stderr)
            print(cmd, file=sys.stderr)
            print('\nThe output was:\n', file=sys.stderr)
            print(error.output.decode(), file=sys.stderr)
            sys.exit(1)

    return True


def decode(x):
    try:
        s = x.decode()
    except:
        return x
    return s
